#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 28 13:25:30 2019

This code will divide a FITS cube into individual chunks.
To minimize problems with how to divide the cube, it will convert the image 
plane into a 1D list of spectra.
Then the file will divided into smaller files, with fewer pixels, in order to
be run through RM synthesis and CLEAN. 
A separate routine will re-assemble the individual chunks into a combined file again.
This code attempts to minimize the memory profile: in principle it should never
need more memory than the size of a single chunk, and perhaps not even that much.

@author: cvaneck
May 2019
"""

import numpy as np
import argparse
import astropy.io.fits as pf
import os.path as path
from math import ceil, floor, log10
#from RMutils.util_misc import progress  

def progress(width, percent):
    """
    Print a progress bar to the terminal.
    Stolen from Mike Bell.
    """
    import math as m
    import sys
    
    marks = m.floor(width * (percent / 100.0))
    spaces = m.floor(width - marks)
    loader = '  [' + ('=' * int(marks)) + (' ' * int(spaces)) + ']'
    sys.stdout.write("%s %d%%\r" % (loader, percent))
    if percent >= 100:
        sys.stdout.write("\n")
    sys.stdout.flush()

def main():
    """This function will divide a large FITS file or cube into smaller chunks.
    It does so in a memory efficient way that requires only a small RAM overhead
    (approximately 1 chunk worth?).
    """
    descStr="""
    Divide a FITS cube into small pieces for memory-efficient RM synthesis.
    Files will be created in running directory.
    WARNING: ONLY WORKS ON FIRST HDU, OTHERS WILL BE LOST."""
    
    parser = argparse.ArgumentParser(description=descStr,
                             formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("infile", metavar="filename.fits", 
                        help="FITS cube containing data")
    parser.add_argument("Nperchunk", metavar="N_pixels", 
                        help="Number of pixels per chunk")
    parser.add_argument("-v", dest="verbose", action="store_true",
                        help="Verbose [False].")
    parser.add_argument("-p", dest="prefix", default=None,
                        help="Prefix of output files [filename]")
        
    args = parser.parse_args()

    Nperchunk=int(args.Nperchunk)

    if not path.exists(args.infile):
        raise Exception('Input file not found!')

    if args.prefix == None:
        prefix=path.splitext(args.infile)[0]
    else:
        prefix=args.prefix




    hdu=pf.open(args.infile,memmap=True)
    header=hdu[0].header
    data=np.transpose(hdu[0].data)
    
    x_image=header['NAXIS1']
    y_image=header['NAXIS2']
    Npix_image=x_image*y_image

    num_chunks=ceil(Npix_image/Nperchunk)
    digits=floor(log10(num_chunks))+1
    prntcode=':0'+str(digits)+'d'
    
    if args.verbose: 
        print(('Chunk name set to "{}.C{'+prntcode+'}.fits"').format(prefix,0))
        print('File will be divided into {} chunks'.format(num_chunks))

    base_idx_arr=np.array(range(Nperchunk))

    new_header=header.copy()
    new_header['NAXIS2']=1
    new_header['NAXIS1']=Nperchunk
    new_header['OLDXDIM']=x_image
    new_header['OLDYDIM']=y_image

    #Run for all but last. Last chunk requires some finessing.
    for i in range(num_chunks-1):
        progress(40,i/num_chunks*100)
        idx=base_idx_arr+i*Nperchunk
        xarr = idx // y_image
        yarr = idx % y_image
        newdata=np.expand_dims(data[xarr,yarr],1)
        filename=('{}.C{'+prntcode+'}.fits').format(prefix,i)
        pf.writeto(filename,np.transpose(newdata),new_header,overwrite=True)

    i+=1
    idx=base_idx_arr+i*Nperchunk
    idx=idx[idx < Npix_image]
    xarr = idx // y_image
    yarr = idx % y_image
    newdata=np.expand_dims(data[xarr,yarr],1)
    filename=('{}.C{'+prntcode+'}.fits').format(prefix,i)
    pf.writeto(filename,np.transpose(newdata),new_header,overwrite=True)

    hdu.close()


if __name__ == "__main__":
    main()
    
    



#!/usr/bin/env python
#=============================================================================#
#                                                                             #
# NAME:     mk_test_ascii_data.py                                             #
#                                                                             #
# PURPOSE:  Create a small ASCII dataset for the purposes of testing the      #
#           RM-code. The script outputs a set ASCII files containing          #
#           frequency and Stokes vectors based on source parameters read from #
#           an input catalogue file. Properties of the output data are set    #
#           in variables at the top of this script. A template describing the #
#           shape of the rms noise curve may also be provided in an external  #
#           ASCII file [freq_Hz, amp].                                        #
#                                                                             #
#           This version for use with the stand-alone code, not the pipeline. #
#                                                                             #
# MODIFIED: 14-mar-2018 by C. Purcell                                         #
#                                                                             #
#=============================================================================#
#                                                                             #
# The MIT License (MIT)                                                       #
#                                                                             #
# Copyright (c) 2015 Cormac R. Purcell                                        #
#                                                                             #
# Permission is hereby granted, free of charge, to any person obtaining a     #
# copy of this software and associated documentation files (the "Software"),  #
# to deal in the Software without restriction, including without limitation   #
# the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# and/or sell copies of the Software, and to permit persons to whom the       #
# Software is furnished to do so, subject to the following conditions:        #
#                                                                             #
# The above copyright notice and this permission notice shall be included in  #
# all copies or substantial portions of the Software.                         #
#                                                                             #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# DEALINGS IN THE SOFTWARE.                                                   #
#                                                                             #
#=============================================================================#

# NOTE: Variations in the noise vs frequency can be specified in an external
# file. Properties of injected sources are given by an external CSV catalogue
# file. Two types of model may be specified, assuming a common flux & spectral
# index. Execute "./0_mk_test_ascii_data.py -h" to print detailed information.

import os
import sys
import argparse
import shutil
import numpy as np

from RMutils.util_misc import create_IQU_spectra_burn
from RMutils.util_misc import create_IQU_spectra_diff
from RMutils.util_misc import csv_read_to_list
from RMutils.util_misc import split_repeat_lst 
from RMutils.util_misc import calc_stats
from RMutils.util_misc import extrap

C = 2.99792458e8


#-----------------------------------------------------------------------------#
def main():
    """
    Run the create_IQU_ascii_data function if called from the command line.
    """
    
    # Help string to be shown using the -h option
    descStr = """
    Create a new dataset directory and populate it with ASCII files containing
    Stokes I, Q and U spectra. Each output file contains four columns
    corresponding to [freq_Hz, StokesI, StokesQ, StokesU] vectors for
    one source.

    The spectra are populated with polarised sources whose properties are given
    in an external CSV-format catalogue file. Two types of model may be 
    specified in the file, assuming a total flux & spectral index:

        # MODEL TYPE 1: One or more components affected by Burn depolarisation.
        # (see e.g., Equation 5 Tribble 1991, Burn 1966).
        # Column  |  Description
        #---------------------------------------------------
        # [0]     |  Model type [1]
        # [1]     |  X coordinate (deg)    ... dummy value
        # [2]     |  Y coordinate (deg)    ... dummy value
        # [3]     |  Major axis (arcsec)   ... dummy value
        # [4]     |  Minor axis (arcsec)   ... dummy value
        # [5]     |  Position angle (deg)  ... dummy value
        # [6]     |  Total flux (mJy)
        # [7]     |  Spectral index
        # Component 1:
        # [8]     |  Intrinsic polarisation angle (deg)
        # [9]     |  Fractional polarisation
        # [10]    |  Faraday depth (radians m^-2)
        # [11]    |  Farday dispersion (radians m^-2)
        # Component 2:
        # [12]    |  Intrinsic polarisation angle (deg)
        # [13]    |  Fractional polarisation
        # [14]    |  Faraday depth (radians m^-2)
        # [15]    |  Farday dispersion (radians m^-2)
        # Component 3:
        # [16]    |  ...
        #---------------------------------------------------

        # MODEL TYPE 2: One or more stacked layers with differential Faraday
        # rotation (Sokoloff et al. 1998, Equation 9).
        #
        # Column  |  Description
        #---------------------------------------------------
        # [0]     |  Model type [2]      
        # [1]     |  X coordinate (deg)    ... dummy value
        # [2]     |  Y coordinate (deg)    ... dummy value
        # [3]     |  Major axis (arcsec)   ... dummy value
        # [4]     |  Minor axis (arcsec)   ... dummy value
        # [5]     |  Position angle (deg)  ... dummy value
        # [6]     |  Total flux (mJy)
        # [7]     |  Spectral index
        # Component 1:
        # [8]     |  Intrinsic polarisation angle (deg)
        # [9]     |  Fractional polarisation
        # [10]    |  Faraday depth (radians m^-2)
        # Component 2:
        # [11]    |  Intrinsic polarisation angle (deg)
        # [12]    |  Fractional polarisation
        # [13]    |  Faraday depth (radians m^-2)
        # Component 3:
        # [14]    |  ...
        #---------------------------------------------------

    Global properties of the data (e.g., frequency sampling, noise level) are
    set by arguments to this script, including an optional template for the
    shape of the noise curve and an optional list of frequency ranges to flag.
    See below for a list and default values in square brackets.

    Examples:

    ./mk_test_ascii_data.py catalogue.csv data/

    ./mk_test_ascii_data.py catalogue.csv -f 1.10e9,1.20e9,1.60e9,1.65e9

    ./mk_test_ascii_data.py catalogue.csv -n NOISE.TXT 

    """

    # Parse the command line options
    parser = argparse.ArgumentParser(description=descStr,
                                 formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("inCatFile", metavar="catalogue.csv", nargs=1,
                        help="Input catalogue file in CSV format")
    parser.add_argument("dataPath", metavar="PATH/TO/DATA",
                        default="data/", nargs="?",
                        help="Path to new data directory [data/]")
    parser.add_argument('-f', nargs=2, metavar="X", type=float,
                        dest='freqRng_MHz', default=[700.0, 1800.0],
                        help='Frequency range [700 1800] (MHz)')
    parser.add_argument("-f0", dest="freq0_MHz", type=float, default=0.0,
                        help="Frequency of catalogue flux [1st channel] (MHz).")
    parser.add_argument("-c", dest="nChans", type=int, default=111,
                        help="Number of channels in output spectra [111].")
    parser.add_argument("-n", dest="rmsNoise", type=float, default=0.02,
                        help="RMS noise of the output spectra [0.02 mJy].")
    parser.add_argument('-t', dest='noiseTmpFile', metavar="NOISE.TXT",
                        help="ASCII file providing a template noise curve (freq amp)")
    parser.add_argument('-l', dest='flagFreqStr', metavar='f1,f2,f1,f2,...',
                        default="", help="Frequency ranges to flag out")
    args = parser.parse_args()
    inCatFile = args.inCatFile[0]
    dataPath = args.dataPath
    startFreq_Hz = args.freqRng_MHz[0] * 1e6
    endFreq_Hz = args.freqRng_MHz[1] * 1e6
    freq0_Hz = None
    if args.freq0_MHz>0.0:
        freq0_Hz = args.freq0_MHz * 1e6
    nChans = args.nChans
    rmsNoise = args.rmsNoise
    noiseTmpFile = args.noiseTmpFile
    flagRanges_Hz = []
    if len(args.flagFreqStr)>0:
        try:
            flagFreqLst = args.flagFreqStr.split(",")
            flagFreqLst = [float(x) for x in flagFreqLst]
            flagRanges_Hz = list(zip(*[iter(flagFreqLst)]*2))
        except Exception:
            "Warn: Failed to parse frequency flagging string!"

    # Read the RMS noise template
    try:
        noiseTmpArr = np.loadtxt(noiseTmpFile, unpack=True)
    except Exception:
        noiseTmpArr = None
    
    # Call the function to create the ASCII data files on disk
    nSrc = create_IQU_ascii_data(dataPath, inCatFile, startFreq_Hz,
                                 endFreq_Hz, nChans, rmsNoise,
                                 noiseTmpArr, flagRanges_Hz, freq0_Hz)

    
#-----------------------------------------------------------------------------#
def create_IQU_ascii_data(dataPath, inCatFile, startFreq_Hz, endFreq_Hz, 
                          nChans, rmsNoise, noiseTmpArr=None,
                          flagRanges_Hz=[], freq0_Hz=None):
    """
    Create a set of ASCII files containing Stokes I Q & U spectra.
    """
    
    # Sample frequency space
    freqArr_Hz = np.linspace(startFreq_Hz, endFreq_Hz, nChans)
    freqNoFlgArr_Hz = freqArr_Hz.copy()
    dFreq_Hz = (endFreq_Hz - startFreq_Hz)/ (nChans-1)
    print("\nSampling frequencies %.2f - %.2f MHz by %.2f MHz." % \
          (freqArr_Hz[0]/1e6, freqArr_Hz[-1]/1e6, dFreq_Hz/1e6))
    if len(flagRanges_Hz)>0:
        print("Flagging frequency ranges:")
        print("> ", flagRanges_Hz)
    for i in range(len(freqArr_Hz)):
        for fRng in flagRanges_Hz:            
            if freqArr_Hz[i]>=fRng[0] and freqArr_Hz[i]<=fRng[1]:
                freqArr_Hz[i]=np.nan

    # Create normalised noise array from a template or assume all ones.
    print("Input RMS noise is %.3g mJy" % rmsNoise)
    if noiseTmpArr is None:
        print("Assuming flat noise versus frequency curve.")
        noiseArr = np.ones(freqArr_Hz.shape, dtype="f8")
    else:
        print("Scaling noise by external template curve.")
        xp = noiseTmpArr[0]
        yp = noiseTmpArr[1]
        mDict = calc_stats(yp)
        yp /= mDict["median"]
        noiseArr = extrap(freqArr_Hz, xp, yp)
        
    # Check the catalogue file exists
    if not os.path.exists(inCatFile):
        print("Err: File does not exist '%s'." % inCatFile)
        sys.exit()
    catInLst = csv_read_to_list(inCatFile, doFloat=True)
    print("Found %d entries in the catalogue." % len(catInLst))

    # Set the frequency at which the flux has been defined
    if freq0_Hz is None:
        freq0_Hz = startFreq_Hz
    print("Catalogue flux is defined at %.3f MHz" % (freq0_Hz/1e6))
    
    # Create the output directory path
    dataPath = dataPath.rstrip("/")
    print("Creating test dataset in '%s/'" % dataPath)
    dirs = dataPath.split("/")
    for i in range(1, len(dirs)):
        dirStr = "/".join(dirs[:i])
        if not os.path.exists(dirStr):
            os.mkdir(dirStr)
    if os.path.exists(dataPath):
        print("\n ", end=' ')
        print("*** WARNING ***" *5)
        print("  About to delete existing data directory!", end=' ')
        print("Previous results will be deleted.\n")
        print("Press <RETURN> to continue ...", end=' ')
        input()
        shutil.rmtree(dataPath, True)
    os.mkdir(dataPath)

    # Loop through the sources, calculate the spectra and save to disk
    successCount = 0
    for i in range(len(catInLst)):
        e = catInLst[i]
        modelType = int(e[0])

        # Type 1 = multiple Burn depolarisation affected components
        if modelType==1:
            
            # Parse the parameters of multiple components
            preLst, parmArr = split_repeat_lst(e[1:],7,4)
            
            # Create the model spectra from multiple thin components
            # modified by external depolarisation
            IArr, QArr, UArr = \
                create_IQU_spectra_burn(freqArr_Hz = freqArr_Hz,
                                        fluxI = preLst[5], 
                                        SI = preLst[6],
                                        fracPolArr = parmArr[0],
                                        psi0Arr_deg = parmArr[1],
                                        RMArr_radm2 = parmArr[2],
                                        sigmaRMArr_radm2 = parmArr[3],
                                        freq0_Hz = freq0_Hz)
                
        # Type 2 = multiple internal depolarisation affected components
        elif modelType==2:
            
            # Parse the parameters of multiple components
            preLst, parmArr = split_repeat_lst(e[1:],7,3)
            
            # Create the model spectra from multiple components
            # modified by internal Faraday depolarisation
            IArr, QArr, UArr = \
                create_IQU_spectra_diff(freqArr_Hz = freqArr_Hz,
                                        fluxI = preLst[5], 
                                        SI = preLst[6],
                                        fracPolArr = parmArr[0],
                                        psi0Arr_deg = parmArr[1],
                                        RMArr_radm2 = parmArr[2],
                                        freq0_Hz = freq0_Hz)
        else:
            continue
        
        # Add scatter to the data to simulate noise
        IArr += (np.random.normal(scale=rmsNoise, size=IArr.shape)
                    * noiseArr)
        QArr += (np.random.normal(scale=rmsNoise, size=QArr.shape)
                    * noiseArr)
        UArr += (np.random.normal(scale=rmsNoise, size=UArr.shape)
                    * noiseArr)
        dIArr = noiseArr * rmsNoise
        dIArr *= np.random.normal(loc=1.0, scale=0.05, size=noiseArr.shape)
        dQArr = noiseArr * rmsNoise 
        dQArr *= np.random.normal(loc=1.0, scale=0.05, size=noiseArr.shape)
        dUArr = noiseArr * rmsNoise 
        dUArr *= np.random.normal(loc=1.0, scale=0.05, size=noiseArr.shape)
        
        # Save spectra to disk
        outFileName = "Source%d.dat" % (i+1)
        outFilePath = dataPath + "/" + outFileName
        print("> Writing ASCII file '%s' ..." % outFileName, end=' ')
        np.savetxt(outFilePath,
                   np.column_stack((freqArr_Hz,
                                    IArr,
                                    QArr,
                                    UArr,
                                    dIArr,
                                    dQArr,
                                    dUArr)))
        print("done.")
        successCount += 1

    return successCount


#-----------------------------------------------------------------------------#
if __name__=="__main__":
    main()



#!/usr/bin/env python
#=============================================================================#
#                                                                             #
# NAME:     do_fitIcube.py                                                    #
#                                                                             #
# PURPOSE:  Make a model Stokes I cube and a noise vector.                    #
#                                                                             #
# MODIFIED: 26-Feb-2017 by C. Purcell                                         #
#                                                                             #
#=============================================================================#
#                                                                             #
# The MIT License (MIT)                                                       #
#                                                                             #
# Copyright (c) 2016 Cormac R. Purcell                                        #
#                                                                             #
# Permission is hereby granted, free of charge, to any person obtaining a     #
# copy of this software and associated documentation files (the "Software"),  #
# to deal in the Software without restriction, including without limitation   #
# the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# and/or sell copies of the Software, and to permit persons to whom the       #
# Software is furnished to do so, subject to the following conditions:        #
#                                                                             #
# The above copyright notice and this permission notice shall be included in  #
# all copies or substantial portions of the Software.                         #
#                                                                             #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# DEALINGS IN THE SOFTWARE.                                                   #
#                                                                             #
#=============================================================================#

import sys
import os
import time
import traceback
import argparse
import math as m
import numpy as np
import astropy.io.fits as pf

from RMutils.util_misc import MAD
from RMutils.util_misc import fit_spec_poly5
from RMutils.util_misc import poly5
from RMutils.util_misc import progress
from RMutils.util_FITS import strip_fits_dims

#-----------------------------------------------------------------------------#
def main():
    """
    Start the make_model_I function if called from the command line.
    """

    # Help string to be shown using the -h option
    descStr = """
    Create a model Stokes I dataset by fitting a polynomial to emitting regions
    above a cutoff in the Stokes I cube.

    NOTE: This script is very simple and designed to only produce a basic model
    Stokes I cube and noise spectrum. In production environments you should
    use a model cube from a full-featured source finder. Treat the output of
    this script as an example of the data format required by the later 
    RM-synthesis script.
    """

    # Parse the command line options
    parser = argparse.ArgumentParser(description=descStr,
                                 formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("fitsI", metavar="StokesI.fits", nargs=1,
                        help="FITS cube containing Stokes I data.")
    parser.add_argument("freqFile", metavar="freqs_Hz.dat", nargs=1,
                        help="ASCII file containing the frequency vector.")
    parser.add_argument("-p", dest="polyOrd", type=int, default=3,
                        help="polynomial order to fit to I spectrum [3].")
    parser.add_argument("-c", dest="cutoff", type=float, default=-3,
                        help="emission cutoff (+ve = abs, -ve = sigma) [-3].")
    parser.add_argument("-o", dest="prefixOut", default="",
                        help="Prefix to prepend to output files [None].")
    parser.add_argument("-b", dest="buffCols", type=int, default=10,
                        help="# pixel columns to buffer for disk IO [10].")
    parser.add_argument("-D", dest="debug", action="store_true",
                        help="turn on debugging messages [False].")
    args = parser.parse_args()
    
    # Sanity checks
    for f in [args.fitsI[0], args.freqFile[0]]:
        if not os.path.exists(f):
            print("File does not exist: '%s'." % f)
            sys.exit()
    dataDir, dummy = os.path.split(args.fitsI[0])

    # Run RM-synthesis on the spectra
    make_model_I(fitsI = args.fitsI[0],
                 freqFile = args.freqFile[0],
                 polyOrd      = args.polyOrd,
                 cutoff       = args.cutoff,
                 prefixOut    = args.prefixOut,
                 outDir       = dataDir,
                 debug        = args.debug,
                 buffCols     = args.buffCols)


#-----------------------------------------------------------------------------#
def make_model_I(fitsI, freqFile, polyOrd=3, cutoff=-1, prefixOut="",
                 outDir="", debug=True, verbose=True, buffCols=10):
    """
    Detect emission in a cube and fit a polynomial model spectrum to the
    emitting pixels. Create a representative noise spectrum using the residual
    planes.
    """

    # Default data type
    
    # Sanity check on header dimensions
    print("Reading FITS cube header from '%s':" % fitsI)
    headI = pf.getheader(fitsI, 0)
    nDim = headI["NAXIS"]
    if nDim < 3 or nDim > 4:
        print("Err: only 3 or 4 dimensions supported: D = %d." % headI["NAXIS"])
        sys.exit()
    nDim = headI["NAXIS"]

    #Idenfitify frequency axis:
    freq_axis=0 #Default for 'frequency axis not identified'
    #Check for frequency axes. Because I don't know what different formatting
    #I might get ('FREQ' vs 'OBSFREQ' vs 'Freq' vs 'Frequency'), convert to
    #all caps and check for 'FREQ' anywhere in the axis name. Shouldn't be more
    #than one axis like that, right?
    for i in range(1,nDim+1):
        try:
            if 'FREQ' in headI['CTYPE'+str(i)].upper():
                freq_axis=i
        except:
            pass #The try statement is needed for if the FITS header does not
                 # have CTYPE keywords.

    nBits=np.abs(headI['BITPIX'])    
    dtFloat = "float" + str(nBits)

    
    nChan = headI["NAXIS"+str(freq_axis)]
    
    # Read the frequency vector
    print("Reading frequency vector from '%s'." % freqFile)
    freqArr_Hz = np.loadtxt(freqFile, dtype=dtFloat)
    freqArr_GHz = freqArr_Hz/1e9
    if nChan!=len(freqArr_Hz):
        print("Err: frequency vector and axis 3 of cube unequal length.")
        sys.exit()
        
    # Measure the RMS spectrum using 2 passes of MAD on each plane
    # Determine which pixels have emission above the cutoff
    print("Measuring the RMS noise and creating an emission mask")
    rmsArr = np.zeros_like(freqArr_Hz)
    mskSrc = np.zeros((headI["NAXIS2"], headI["NAXIS1"]), dtype=dtFloat)
    mskSky = np.zeros((headI["NAXIS2"], headI["NAXIS1"]), dtype=dtFloat)
    for i in range(nChan):
        HDULst = pf.open(fitsI, "readonly", memmap=True)
        if nDim==3:
            dataPlane = HDULst[0].data[i,:,:]
        elif nDim==4 and freq_axis == 4:
            dataPlane = HDULst[0].data[i,0,:,:]
        elif nDim==4 and freq_axis == 3:
            dataPlane = HDULst[0].data[0,i,:,:]
        if cutoff>0:
            idxSky = np.where(dataPlane<cutoff)
        else:
            idxSky = np.where(dataPlane)
        
        # Pass 1
        rmsTmp = MAD(dataPlane[idxSky])
        medTmp = np.nanmedian(dataPlane[idxSky])
        
        # Pass 2: use a fixed 3-sigma cutoff to mask off emission
        idxSky = np.where(dataPlane < medTmp + rmsTmp * 3)
        medSky = np.nanmedian(dataPlane[idxSky])
        rmsArr[i] = MAD(dataPlane[idxSky])
        mskSky[idxSky] +=1
        
        # When building final emission mask treat +ve cutoffs as absolute
        # values and negative cutoffs as sigma values
        if cutoff>0:
            idxSrc = np.where(dataPlane > cutoff)
        else:
            idxSrc = np.where(dataPlane > medSky -1 * rmsArr[i] * cutoff)
        mskSrc[idxSrc] +=1

        # Clean up
        HDULst.close()
        del HDULst

    # Save the noise spectrum
    if outDir == '':
        outDir='.'
    print("Saving the RMS noise spectrum in an ASCII file:")
    outFile = outDir + "/"  + prefixOut + "Inoise.dat"
    print("> %s" % outFile)
    np.savetxt(outFile, rmsArr)
        
    # Save FITS files containing sky and source masks
    print("Saving sky and source mask images:")
    mskArr = np.where(mskSky>0, 1.0, np.nan)
    headMsk = strip_fits_dims(header=headI, minDim=2)
    headMsk["DATAMAX"] = 1
    headMsk["DATAMIN"] = 0
    del headMsk["BUNIT"]
    fitsFileOut = outDir + "/"  + prefixOut + "IskyMask.fits"
    print("> %s" % fitsFileOut)
    pf.writeto(fitsFileOut, mskArr, headMsk, output_verify="fix",
               overwrite=True)
    mskArr = np.where(mskSrc>0, 1.0, np.nan)
    fitsFileOut = outDir + "/"  + prefixOut + "IsrcMask.fits"
    print("> %s" % fitsFileOut)
    pf.writeto(fitsFileOut, mskArr, headMsk, output_verify="fix",
               overwrite=True)
        
    # Create a blank FITS file on disk using the large file method
    # http://docs.astropy.org/en/stable/io/fits/appendix/faq.html
    #  #how-can-i-create-a-very-large-fits-file-from-scratch
    fitsModelFile = outDir + "/"  + prefixOut + "Imodel.fits"
    print("Creating an empty FITS file on disk")
    print("> %s" % fitsModelFile)
    stub = np.zeros((10, 10, 10), dtype=dtFloat)
    hdu = pf.PrimaryHDU(data=stub)
    headModel = strip_fits_dims(header=headI, minDim=nDim)
    headModel["NAXIS1"] = headI["NAXIS1"]
    headModel["NAXIS2"] = headI["NAXIS2"]
    headModel["NAXIS3"] = headI["NAXIS3"]
    nVoxels = headI["NAXIS1"] * headI["NAXIS2"] * headI["NAXIS3"]
    if nDim==4:
        headModel["NAXIS4"] = headI["NAXIS4"]
        nVoxels *= headI["NAXIS4"]
    while len(headModel) < (36 * 4 - 1):
        headModel.append()
    headModel.tofile(fitsModelFile, overwrite=True)
    with open(fitsModelFile, "rb+") as f:
        f.seek(len(headModel.tostring()) + (nVoxels*int(nBits/8)) - 1)
        f.write(b"\0")

    # Feeback to user
    srcIdx = np.where(mskSrc>0)
    srcCoords = np.rot90(np.where(mskSrc>0))
    if verbose:
        nPix = mskSrc.shape[-1] * mskSrc.shape[-2]
        nDetectPix = len(srcCoords)
        print("Emission present in %d spectra (%.1f percent)." % \
              (nDetectPix, (nDetectPix*100.0/nPix)))

    # Inform user job magnitude
    startTime = time.time()
    print("Fitting %d/%d spectra." % (nDetectPix, nPix))
    j = 0
    nFailPix = 0
    if verbose:
        progress(40, 0)
        
    # Loop through columns of pixels (buffers disk IO)
    for i in range(0, headI["NAXIS1"], buffCols):

        # Select the relevant pixel columns from the mask and cube
        mskSub = mskSrc[:, i:i+buffCols]
        srcCoords = np.rot90(np.where(mskSub>0))
        
        # Select the relevant pixel columns from the mask
        HDULst = pf.open(fitsI, "readonly", memmap=True)
        if nDim==3:
            IArr = HDULst[0].data[:, :, i:i+buffCols]
        elif nDim==4 and freq_axis == 3:
            IArr = HDULst[0].data[0,:, :, i:i+buffCols]
        elif nDim==4 and freq_axis == 4:
            IArr = HDULst[0].data[:,0, :, i:i+buffCols]

        HDULst.close()
        IModArr = np.ones_like(IArr,dtype=dtFloat)*medSky

        # Fit the spectra in turn
        for yi, xi in srcCoords:
            j += 1
            if verbose:
                progress(40, ((j)*100.0/nDetectPix))
                
            # Fit a <=5th order polynomial model to the Stokes I spectrum
            # Frequency axis must be in GHz to avoid overflow errors
            fitDict = {"fitStatus": 0,
                       "chiSq": 0.0,
                       "dof": len(freqArr_GHz)-polyOrd-1,
                       "chiSqRed": 0.0,
                       "nIter": 0,
                       "p": None}
            try:
                mp = fit_spec_poly5(freqArr_GHz,
                                    IArr[:,yi,xi],
                                    rmsArr,
                                    polyOrd)
                fitDict["p"]         = mp.params
                fitDict["fitStatus"] = mp.status
                fitDict["chiSq"]     = mp.fnorm
                fitDict["chiSqRed"]  = mp.fnorm/fitDict["dof"]
                fitDict["nIter"]     = mp.niter
                IModArr[:,yi,xi]       = poly5(fitDict["p"])(freqArr_GHz)
            
            except Exception:
                nFailPix += 1
                if debug:
                    print("\nTRACEBACK:")
                    print("-" * 80)
                    print(traceback.format_exc())
                    print("-" * 80)
                    print() 
                    print("> Setting Stokes I spectrum to NaN.\n")
                fitDict["p"] = [0.0, 0.0, 0.0, 0.0, 0.0, 1.0]
                IModArr[:,yi,xi] = np.ones_like(IArr[:,yi,xi]) * np.nan
        
        # Write the spectrum to the model file
        HDULst = pf.open(fitsModelFile, "update", memmap=True)
        if nDim==3:            
            HDULst[0].data[:, :, i:i+buffCols] = IModArr
        elif nDim==4 and freq_axis == 3:
            HDULst[0].data[0, :, :, i:i+buffCols] = IModArr
        elif nDim==4 and freq_axis == 4:
            HDULst[0].data[:, 0, :, i:i+buffCols] = IModArr
        HDULst.close()
        
    endTime = time.time()
    cputime = (endTime - startTime)
    print("Fitting completed in %.2f seconds." % cputime)
    if nFailPix>0:
        print("Warn: Fitting failed on %d/%d spectra  (%.1f percent)." % \
              (nFailPix, nDetectPix, (nFailPix*100.0/nDetectPix)))
    
    
#-----------------------------------------------------------------------------#
if __name__ == "__main__":
    main()



#!/usr/bin/env python
#=============================================================================#
#                                                                             #
# NAME:     do_RMclean_1D.py                                                  #
#                                                                             #
# PURPOSE:  Command line functions for RM-clean                               #
#           on a dirty Faraday dispersion function.                           #
# CREATED:  16-Nov-2018 by J. West                                            #
# MODIFIED: 16-Nov-2018 by J. West                                            #
# MODIFIED: 23-October-2019 by A. Thomson                                     #
#                                                                             #
#=============================================================================#
#                                                                             #
# The MIT License (MIT)                                                       #
#                                                                             #
# Copyright (c) 2015 Cormac R. Purcell                                        #
#                                                                             #
# Permission is hereby granted, free of charge, to any person obtaining a     #
# copy of this software and associated documentation files (the "Software"),  #
# to deal in the Software without restriction, including without limitation   #
# the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# and/or sell copies of the Software, and to permit persons to whom the       #
# Software is furnished to do so, subject to the following conditions:        #
#                                                                             #
# The above copyright notice and this permission notice shall be included in  #
# all copies or substantial portions of the Software.                         #
#                                                                             #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# DEALINGS IN THE SOFTWARE.                                                   #
#                                                                             #
#=============================================================================#

import time
import os
import sys
import json
import numpy as np

from RMutils.util_RM import do_rmclean_hogbom
from RMutils.util_RM import measure_FDF_parms
from RMutils.util_RM import measure_fdf_complexity

C = 2.997924538e8 # Speed of light [m/s]
#-----------------------------------------------------------------------------#
def run_rmclean(mDictS, aDict, cutoff,
                maxIter=1000, gain=0.1, nBits=32,
                showPlots=False, verbose=False, log=print):
    """Run RM-CLEAN on a complex FDF spectrum given a RMSF.

    Args:
        mDictS (dict): Summary of RM synthesis results.
        aDict (dict): Data output by RM synthesis.
        cutoff (float): CLEAN cutoff in flux units

    Kwargs:
        maxIter (int): Maximum number of CLEAN iterations per pixel.
        gain (float): CLEAN loop gain.
        nBits (int): Precision of floating point numbers.
        showPlots (bool): Show plots?
        verbose (bool): Verbosity.
        log (function): Which logging function to use.

    Returns:
        mDict (dict): Summary of RMCLEAN results.
        arrdict (dict): Data output by RMCLEAN.

    """
    phiArr_radm2 = aDict["phiArr_radm2"]
    freqArr_Hz = aDict["freqArr_Hz"]
    weightArr = aDict["weightArr"]
    dirtyFDF = aDict["dirtyFDF"]
    phi2Arr_radm2 = aDict["phi2Arr_radm2"]
    RMSFArr=aDict["RMSFArr"]


    lambdaSqArr_m2 = np.power(C/freqArr_Hz, 2.0)

    # If the cutoff is negative, assume it is a sigma level
    if verbose: log("Expected RMS noise = %.4g flux units" % (mDictS["dFDFth"]))
    if cutoff<0:
        if verbose: log("Using a sigma cutoff of %.1f." %  (-1 * cutoff))
        cutoff = -1 * mDictS["dFDFth"] * cutoff
        if verbose: log("Absolute value = %.3g" % cutoff)
    else:
        if verbose: log("Using an absolute cutoff of %.3g (%.1f x expected RMS)." % (cutoff, cutoff/mDictS["dFDFth"]))

    startTime = time.time()
    # Perform RM-clean on the spectrum
    cleanFDF, ccArr, iterCountArr, residFDF = \
              do_rmclean_hogbom(dirtyFDF        = dirtyFDF,
                                phiArr_radm2    = phiArr_radm2,
                                RMSFArr         = RMSFArr,
                                phi2Arr_radm2   = phi2Arr_radm2,
                                fwhmRMSFArr     = np.array(mDictS["fwhmRMSF"]),
                                cutoff          = cutoff,
                                maxIter         = maxIter,
                                gain            = gain,
                                verbose         = verbose,
                                doPlots         = showPlots)

    # ALTERNATIVE RM_CLEAN CODE ----------------------------------------------#
    '''
    cleanFDF, ccArr, fwhmRMSF, iterCount = \
              do_rmclean(dirtyFDF     = dirtyFDF,
                         phiArr       = phiArr_radm2,
                         lamSqArr     = lamSqArr_m2,
                         cutoff       = cutoff,
                         maxIter      = maxIter,
                         gain         = gain,
                         weight       = weightArr,
                         RMSFArr      = RMSFArr,
                         RMSFphiArr   = phi2Arr_radm2,
                         fwhmRMSF     = mDictS["fwhmRMSF"],
                         doPlots      = True)
    '''
    #-------------------------------------------------------------------------#

    endTime = time.time()
    cputime = (endTime - startTime)
    if verbose: log("> RM-CLEAN completed in %.4f seconds." % cputime)

    # Measure the parameters of the deconvolved FDF
    mDict = measure_FDF_parms(FDF         = cleanFDF,
                              phiArr      = phiArr_radm2,
                              fwhmRMSF    = mDictS["fwhmRMSF"],
                              dFDF        = mDictS["dFDFth"],
                              lamSqArr_m2 = lambdaSqArr_m2,
                              lam0Sq      = mDictS["lam0Sq_m2"])
    mDict["cleanCutoff"] = cutoff
    mDict["nIter"] = int(iterCountArr)

    # Measure the complexity of the clean component spectrum
    mDict["mom2CCFDF"] = measure_fdf_complexity(phiArr = phiArr_radm2,
                                                FDF = ccArr)

    #Calculating observed errors (based on dFDFcorMAD)
    mDict["dPhiObserved_rm2"] = mDict["dPhiPeakPIfit_rm2"]*mDict["dFDFcorMAD"]/mDictS["dFDFth"]
    mDict["dAmpObserved"] = mDict["dFDFcorMAD"]
    mDict["dPolAngleFitObserved_deg"] = mDict["dPolAngleFit_deg"]*mDict["dFDFcorMAD"]/mDictS["dFDFth"]

    mDict["dPolAngleFit0Observed_deg"] = mDict["dPolAngle0Fit_deg"]*mDict["dFDFcorMAD"]/mDictS["dFDFth"]





    if verbose:
        # Print the results to the screen
        log()
        log('-'*80)
        log('RESULTS:\n')
        log('FWHM RMSF = %.4g rad/m^2' % (mDictS["fwhmRMSF"]))
        log('Pol Angle = %.4g (+/-%.4g observed, +- %.4g theoretical) deg' % 
            (mDict["polAngleFit_deg"],mDict["dPolAngleFitObserved_deg"],mDict["dPolAngleFit_deg"]))
        log('Pol Angle 0 = %.4g (+/-%.4g observed, +- %.4g theoretical) deg' % 
            (mDict["polAngle0Fit_deg"],mDict["dPolAngle0ChanObserved_deg"],mDict["dPolAngle0Fit_deg"]))
        log('Peak FD = %.4g (+/-%.4g observed, +- %.4g theoretical) rad/m^2' % 
            (mDict["phiPeakPIfit_rm2"],mDict["dPhiObserved_rm2"],mDict["dPhiPeakPIfit_rm2"]))
        log('freq0_GHz = %.4g ' % (mDictS["freq0_Hz"]/1e9))
        log('I freq0 = %.4g %s' % (mDictS["Ifreq0"],mDictS["units"]))
        log('Peak PI = %.4g (+/-%.4g observed, +- %.4g theoretical) %s' % 
            (mDict["ampPeakPIfit"],mDict["dAmpObserved"],mDict["dAmpPeakPIfit"],mDictS["units"]))
        log('QU Noise = %.4g %s' % (mDictS["dQU"],mDictS["units"]))
        log('FDF Noise (theory)   = %.4g %s' % (mDictS["dFDFth"],mDictS["units"]))
        log('FDF Noise (Corrected MAD) = %.4g %s' % (mDict["dFDFcorMAD"],mDictS["units"]))
        log('FDF Noise (rms)   = %.4g %s' % (mDict["dFDFrms"],mDictS["units"]))

        log('FDF SNR = %.4g ' % (mDict["snrPIfit"]))
        log()
        log('-'*80)

    # Pause to display the figure
    if showPlots:
        plot_clean_spec(phiArr_radm2, dirtyFDF, cleanFDF, ccArr, residFDF,
                    cutoff,mDictS["units"])
        print("Press <RETURN> to exit ...", end=' ')
        input()

    #add array dictionary
    arrdict = dict()
    arrdict["phiArr_radm2"] = phiArr_radm2
    arrdict["freqArr_Hz"] = freqArr_Hz
    arrdict["cleanFDF"]=cleanFDF
    arrdict["ccArr"]=ccArr
    arrdict["iterCountArr"]=iterCountArr
    arrdict["residFDF"]=residFDF

    return mDict, arrdict


def saveOutput(outdict, arrdict, prefixOut="", verbose=False, log=print):
    """
    Saves RM-CLEAN results to text files. The clean (restored) FDF, model FDF 
    (clean components) are saved, as is two files (.dat and .json)
    reporting the fitting results as key-value pairs.
    Inputs:
        outdict: the results dictionary (mDict) from RM-CLEAN.
        arrdict: the array dictionary (aDict) from RM-CLEAN.
        prefixOut (str): name prefix to be given to output files 
            (including relative/absolute directory to save to)
        verbose (bool): print verbose messages?
        log (function): function to use when printing verbose messages.
    """
    # Get data
    phiArr_radm2 = arrdict["phiArr_radm2"]
    cleanFDF = arrdict["cleanFDF"]
    ccArr = arrdict["ccArr"]

    # Save the deconvolved FDF and CC model to ASCII files
    if verbose: log("Saving the clean FDF and component model to ASCII files.")
    outFile = prefixOut + "_FDFclean.dat"
    if verbose: log("> %s" % outFile)
    np.savetxt(outFile, list(zip(phiArr_radm2, cleanFDF.real, cleanFDF.imag)))
    outFile = prefixOut + "_FDFmodel.dat"
    if verbose: log("> %s" % outFile)
    np.savetxt(outFile, list(zip(phiArr_radm2, ccArr.real, ccArr.imag)))

    # Save the RM-clean measurements to a "key=value" text file
    if verbose: log("Saving the measurements on the FDF in 'key=val' and JSON formats.")
    outFile = prefixOut + "_RMclean.dat"
    if verbose: log("> %s" % outFile)
    FH = open(outFile, "w")
    for k, v in outdict.items():
        FH.write("%s=%s\n" % (k, v))
    FH.close()
    outFile = prefixOut + "_RMclean.json"
    if verbose: log("> %s" % outFile)
    json.dump(outdict, open(outFile, "w"))

def readFiles(fdfFile, rmsfFile, weightFile, rmSynthFile, nBits):
    """
    Read in the RM-synthesis output files and assemble back into dictionaries.
    Inputs:
        fdfFile (str): file path to the FDF.
        rmsfFile (str): file path to the RMSF.
        weightfile (str): file path to the channel weights.
        rmSynthFile (str): file path to the RMsynth json file.
        nBits (int): number of bits to use when storing the data.
    """

    # Default data types
    dtFloat = "float" + str(nBits)
    dtComplex = "complex" + str(2*nBits)

    # Read the RMSF from the ASCII file
    phi2Arr_radm2, RMSFreal, RMSFimag = np.loadtxt(rmsfFile, unpack=True, dtype=dtFloat)
    # Read the frequency vector for the lambda^2 array
    freqArr_Hz, weightArr = np.loadtxt(weightFile, unpack=True, dtype=dtFloat)
    # Read the FDF from the ASCII file
    phiArr_radm2, FDFreal, FDFimag = np.loadtxt(fdfFile, unpack=True, dtype=dtFloat)
    # Read the RM-synthesis parameters from the JSON file
    mDictS = json.load(open(rmSynthFile, "r"))
    dirtyFDF = FDFreal + 1j * FDFimag
    RMSFArr = RMSFreal + 1j * RMSFimag

    #add array dictionary
    aDict = dict()
    aDict["phiArr_radm2"] = phiArr_radm2
    aDict["phi2Arr_radm2"] = phi2Arr_radm2
    aDict["RMSFArr"] = RMSFArr
    aDict["freqArr_Hz"] = freqArr_Hz
    aDict["weightArr"]=weightArr
    aDict["dirtyFDF"]=dirtyFDF

    return mDictS, aDict


def plot_clean_spec(phiArr_radm2, dirtyFDF, cleanFDF, ccArr, residFDF,
                    cutoff,units):
    """
    Plotting code for CLEANed Faraday depth spectra.
    Inputs: 
        phiArr_radm2 (array): array of Faraday depth values.
        dirty FDF (array): dirty Faraday depth spectrum.
        cleanFDF (array): cleaned (restored) Faraday depth spectrum
        ccArr (array): clean component array
        residFDF (array): residual Faraday depth spectrum
        cutoff (float): clean threshold
        units (str): name of flux unit
    """
    from matplotlib import pyplot as plt
    from matplotlib.ticker import MaxNLocator

    fig = plt.figure(figsize=(12.0, 8))
    ax1 = fig.add_subplot(211)
    ax2 = fig.add_subplot(212, sharex=ax1)

    dirtyFDF=np.squeeze(dirtyFDF)
    cleanFDF=np.squeeze(cleanFDF)
    ccArr=np.squeeze(ccArr)
    residFDF=np.squeeze(residFDF)

    fig.show()


    ax1.cla()
    ax2.cla()
    ax1.step(phiArr_radm2, np.abs(dirtyFDF),
             color="grey", marker="None", mfc="w", mec="g", ms=10,
             where="mid", label="Dirty FDF")
    ax1.step(phiArr_radm2, np.abs(ccArr), color="g",
             marker="None", mfc="w", mec="g", ms=10, where="mid",
             label="Clean Components")
    ax1.step(phiArr_radm2, np.abs(residFDF), color="magenta",
             marker="None", mfc="w", mec="g", ms=10, where="mid",
             label="Residual FDF")
    ax1.step(phiArr_radm2, np.abs(cleanFDF), color="k",
             marker="None", mfc="w", mec="g", ms=10, where="mid", lw=1.5,
             label="Clean FDF")
    ax1.axhline(cutoff, color="r", ls="--", label="Clean cutoff")
    ax1.yaxis.set_major_locator(MaxNLocator(4))
    ax1.set_ylabel("Flux Density ("+units+')')
    leg = ax1.legend(numpoints=1, loc='upper right', shadow=False,
                     borderaxespad=0.3, bbox_to_anchor=(1.00, 1.00))
    for t in leg.get_texts():
        t.set_fontsize('small')
    leg.get_frame().set_linewidth(0.5)
    leg.get_frame().set_alpha(0.5)
    [label.set_visible(False) for label in ax1.get_xticklabels()]
    ax2.step(phiArr_radm2, np.abs(residFDF), color="magenta",
             marker="None", mfc="w", mec="g", ms=10, where="mid",
             label="Residual FDF")
    ax2.step(phiArr_radm2, np.abs(ccArr), color="g",
             marker="None", mfc="w", mec="g", ms=10, where="mid",
             label="Clean Components")
    ax2.axhline(cutoff, color="r", ls="--", label="Clean cutoff")
    ax2.set_ylim(0, cutoff*3.0)
    ax2.yaxis.set_major_locator(MaxNLocator(4))
    ax2.set_ylabel("Flux Density ("+units+')')
    ax2.set_xlabel("$\phi$ rad m$^{-2}$")
    leg = ax2.legend(numpoints=1, loc='upper right', shadow=False,
                     borderaxespad=0.3, bbox_to_anchor=(1.00, 1.00))
    for t in leg.get_texts():
        t.set_fontsize('small')
    leg.get_frame().set_linewidth(0.5)
    leg.get_frame().set_alpha(0.5)
    ax2.autoscale_view(True, True, True)
    plt.draw()

#-----------------------------------------------------------------------------#
def main():
    import argparse
    """
    Start the function to perform RM-clean if called from the command line.
    """

    # Help string to be shown using the -h option
    descStr = """
    Run RM-CLEAN on an ASCII Faraday dispersion function (FDF), applying
    the rotation measure spread function created by the script
    'do_RMsynth_1D.py'. Saves ASCII files containing a deconvolved FDF &
    clean-component spectrum.
    """

    epilog_text="""
    By default, saves the following files:
    _FDFclean.dat: cleaned and restored FDF [Phi, Q, U]
    _FDFmodel.dat: clean component FDF [Phi, Q, U]
    _RMclean.dat: list of calculated paramaters describing FDF
    _RMclean.json: dictionary of calculated parameters
    """

    # Parse the command line options
    parser = argparse.ArgumentParser(description=descStr,epilog=epilog_text,
                                 formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("dataFile", metavar="dataFile.dat", nargs=1,
                        help="ASCII file containing original frequency spectra.")
    parser.add_argument("-c", dest="cutoff", type=float, default=-3,
                        help="CLEAN cutoff (+ve = absolute, -ve = sigma) [-3].")
    parser.add_argument("-n", dest="maxIter", type=int, default=1000,
                        help="maximum number of CLEAN iterations [1000].")
    parser.add_argument("-g", dest="gain", type=float, default=0.1,
                        help="CLEAN loop gain [0.1].")
    parser.add_argument("-p", dest="showPlots", action="store_true",
                        help="show the plots [False].")
    parser.add_argument("-v", dest="verbose", action="store_true",
                        help="Print verbose messages")
    args = parser.parse_args()

    # Form the input file names from prefix of the original data file
    fileRoot, dummy = os.path.splitext(args.dataFile[0])
    fdfFile = fileRoot + "_FDFdirty.dat"
    rmsfFile = fileRoot + "_RMSF.dat"
    weightFile = fileRoot + "_weight.dat"
    rmSynthFile = fileRoot + "_RMsynth.json"
    # Sanity checks
    for f in [weightFile, fdfFile, rmsfFile, rmSynthFile]:
        if not os.path.exists(f):
            print("File does not exist: '{:}'.".format(f), end=' ')
            sys.exit()
    nBits = 32
    dataDir, dummy = os.path.split(args.dataFile[0])
    mDictS, aDict = readFiles(fdfFile, rmsfFile, weightFile, rmSynthFile, nBits)
    # Run RM-CLEAN on the spectrum
    outdict, arrdict = run_rmclean(mDictS  = mDictS,
                                aDict        = aDict,
                                cutoff       = args.cutoff,
                                maxIter      = args.maxIter,
                                gain         = args.gain,
                                nBits        = nBits,
                                showPlots    = args.showPlots,
                                verbose      = args.verbose)

    # Save output
    saveOutput(outdict,
            arrdict,
            prefixOut    = fileRoot,
            verbose      = args.verbose)



#-----------------------------------------------------------------------------#
if __name__ == "__main__":
    main()



#!/usr/bin/env python

#=============================================================================#
#                                                                             #
# NAME:     do_QUfit_1D_nest.py                                               #
#                                                                             #
# PURPOSE:  Code to simultaneously fit Stokes Q/I and U/I spectra with a      #
#           Faraday active models.                                            #
#                                                                             #
# MODIFIED: 12-Mar-2018 by C. Purcell                                         #
#                                                                             #
# CONTENTS:                                                                   #
#                                                                             #
#   main           ... parse command line arguments and initiate procedure    #
#   run_qufit      ... main function of the QU-fitting procedure              #
#   prior_call     ... return the prior tranform fucntion given prior limits  #
#   lnlike_call    ... return a function to evaluate ln(like) given the data  #
#   chisq_model    ... calculate the chi-squared for the model                #
#   wrap_chains    ... wrap and shift chains of periodic parameters           #
#   init_mnest     ... initialise multinest using a default dictionary        #
#   merge_two_dicts .. merge two dictionaries                                 #
#                                                                             #
#=============================================================================#
#                                                                             #
# The MIT License (MIT)                                                       #
#                                                                             #
# Copyright (c) 2018 Cormac R. Purcell                                        #
#                                                                             #
# Permission is hereby granted, free of charge, to any person obtaining a     #
# copy of this software and associated documentation files (the "Software"),  #
# to deal in the Software without restriction, including without limitation   #
# the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# and/or sell copies of the Software, and to permit persons to whom the       #
# Software is furnished to do so, subject to the following conditions:        #
#                                                                             #
# The above copyright notice and this permission notice shall be included in  #
# all copies or substantial portions of the Software.                         #
#                                                                             #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# DEALINGS IN THE SOFTWARE.                                                   #
#                                                                             #
#=============================================================================#

from shutil import Error
import sys
import os
import shutil
import copy
import time
import importlib
import json
import argparse
import traceback
import numpy as np
import matplotlib.pyplot as plt
import pymultinest as pmn
try:
    from mpi4py import MPI
    mpiSwitch = True
except:
    mpiSwitch = False

from RMutils.util_misc import create_frac_spectra
from RMutils.util_misc import poly5
from RMutils.util_misc import toscalar
from RMutils.util_plotTk import plot_Ipqu_spectra_fig
from RMutils.util_plotTk import CustomNavbar
from RMutils import corner

# Fail if script has been started with mpiexec & mpi4py is not installed
if os.environ.get('OMPI_COMM_WORLD_SIZE') is not None:
    if not mpiSwitch:
        print("Script called with mpiexec, but mpi4py not installed")
        sys.exit()
        
C = 2.997924538e8 # Speed of light [m/s]


#-----------------------------------------------------------------------------#
def main():
    """
    Start the run_qufit procedure if called from the command line.
    
    Run QU-fitting on polarised spectra (1D) stored in an ASCII file. The
    Stokes I spectra is first fit with a polynomial and the resulting model
    used to create fractional q = Q/I and u = U/I spectra. If the 'noStokesI'
    option is given, the input data are assumed to be fractional already.

    The script uses the Nested Sampling algorithm (Skilling 2004) to find
    the best fitting parameters, given a prior function on each free parameter.
    The sampling algorithm also calculates the Bayesian evidence, which can
    be used for model comparison. Factors of >10 between models mean that one
    model is strongly favoured over the other.

    Models and priors are  specified as Python code in files called 'mX.py'
    within the 'models_ns' directory. See the existing files for examples
    drawn from the paper Sokoloff et al. 1998, MNRAS 229, pg 189.

    Main algorithm handles the command line interface, passing all arguments
    one to run_qufit().
    """

    # Help string to be shown using the -h option
    descStr = """
    Run QU-fitting on polarised spectra (1D) stored in an ASCII file. The
    Stokes I spectra is first fit with a polynomial and the resulting model
    used to create fractional q = Q/I and u = U/I spectra. If the 'noStokesI'
    option is given, the input data are assumed to be fractional already.

    The script uses the Nested Sampling algorithm (Skilling 2004) to find
    the best fitting parameters, given a prior function on each free parameter.
    The sampling algorithm also calculates the Bayesian evidence, which can
    be used for model comparison. Factors of >10 between models mean that one
    model is strongly favoured over the other.

    Models and priors are  specified as Python code in files called 'mX.py'
    within the 'models_ns' directory. See the existing files for examples
    drawn from the paper Sokoloff et al. 1998, MNRAS 229, pg 189.
    """

    # Parse the command line options
    parser = argparse.ArgumentParser(description=descStr,
                                 formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("dataFile", metavar="dataFile.dat", nargs=1,
                        help="ASCII file containing Stokes spectra & errors.")
    parser.add_argument("-m", dest="modelNum", type=int, default=1,
                        help="model number to fit [1].")
    parser.add_argument("-o", dest="polyOrd", type=int, default=2,
                        help="polynomial order to fit to I spectrum [2].")
    parser.add_argument("-i", dest="noStokesI", action="store_true",
                        help="ignore the Stokes I spectrum [False].")
    parser.add_argument("-p", dest="showPlots", action="store_true",
                        help="show the plots [False].")
    parser.add_argument("-d", dest="debug", action="store_true",
                        help="turn on debugging messages/plots [False].")
    parser.add_argument("-v", dest="verbose", action="store_true",
                        help="verbose mode [False].")
    parser.add_argument("-c", dest="sigma_clip", type=float, default=5,
                        help="How many sigma to clip around parameter modes [5].")
    parser.add_argument("-r", dest="restart", action="store_false",
                        help="Restart fitting and delete previous run [True].")
    args = parser.parse_args()

    # Sanity checks
    if not os.path.exists(args.dataFile[0]):
        print("File does not exist: '%s'." % args.dataFile[0])
        sys.exit()
    dataDir, dummy = os.path.split(args.dataFile[0])

    # Run the QU-fitting procedure
    run_qufit(dataFile     = args.dataFile[0],
              modelNum     = args.modelNum,
              outDir       = dataDir,
              polyOrd      = args.polyOrd,
              nBits        = 32,
              noStokesI    = args.noStokesI,
              showPlots    = args.showPlots,
              sigma_clip   = args.sigma_clip,
              debug        = args.debug,
              verbose      = args.verbose,
              restart      = args.restart
              )


#-----------------------------------------------------------------------------#
def run_qufit(dataFile, modelNum, outDir="", polyOrd=3, nBits=32,
              noStokesI=False, showPlots=False, sigma_clip=5, 
              debug=False, verbose=False, restart=True):
    """Carry out QU-fitting using the supplied parameters:
        dataFile (str, required): relative or absolute path of file containing 
            frequencies and Stokes parameters with errors.
        modelNum (int, required): number of model to be fit to data. Models and
             priors are specified as Python code in files called 'mX.py' within  
            the 'models_ns' directory.
        outDir (str): relative or absolute path to save outputs to. Defaults to
            working directory.
        polyOrd (int): Order of polynomial to fit to Stokes I spectrum (used to
            normalize Q and U values). Defaults to 3 (cubic).
        nBits (int): number of bits to use in internal calculations.
        noStokesI (bool): set True if the Stokes I spectrum should be ignored.
        showPlots (bool): Set true if the spectrum and parameter space plots
            should be displayed.
        sigma_clip (float): How many standard deviations to clip around the 
            mean of each mode in the parameter postierors.
        debug (bool): Display debug messages.
        verbose (bool): Print verbose messages/results to terminal.
        
        Returns: nothing. Results saved to files and/or printed to terminal."""

    # Get the processing environment
    if mpiSwitch:
        mpiComm = MPI.COMM_WORLD
        mpiSize = mpiComm.Get_size()
        mpiRank = mpiComm.Get_rank()
    else:
        mpiSize = 1
        mpiRank = 0
        
    # Default data types
    dtFloat = "float" + str(nBits)
    dtComplex = "complex" + str(2*nBits)

    # Output prefix is derived from the input file name
    prefixOut, ext = os.path.splitext(dataFile)
    nestOut = f"{prefixOut}_m{modelNum}_nest/"
    if mpiRank==0:
        if os.path.exists(nestOut) and restart:
            shutil.rmtree(nestOut, True)
            os.mkdir(nestOut)
        elif not os.path.exists(nestOut) and restart:
            os.mkdir(nestOut)
        elif not os.path.exists(nestOut) and not restart:
            print("Restart requested, but previous run not found!")
            raise Exception(f"{nestOut} does not exist.")
    if mpiSwitch:
        mpiComm.Barrier()

    # Read the data file in the root process
    if mpiRank==0:
        dataArr = np.loadtxt(dataFile, unpack=True, dtype=dtFloat)
    else:
        dataArr = None
    if mpiSwitch:
        dataArr = mpiComm.bcast(dataArr, root=0)

    # Parse the data array
    # freq_Hz, I, Q, U, dI, dQ, dU
    try:
        (freqArr_Hz, IArr, QArr, UArr,
         dIArr, dQArr, dUArr) = dataArr
        if mpiRank==0:
            print("\nFormat [freq_Hz, I, Q, U, dI, dQ, dU]")
    except Exception:
        # freq_Hz, Q, U, dQ, dU
        try:
            (freqArr_Hz, QArr, UArr, dQArr, dUArr) = dataArr
            if mpiRank==0:
                print("\nFormat [freq_Hz, Q, U,  dQ, dU]")
            noStokesI = True
        except Exception:
            print("\nError: Failed to parse data file!")
            if debug:
                print(traceback.format_exc())
            if mpiSwitch:
                MPI.Finalize()
            return

    # If no Stokes I present, create a dummy spectrum = unity
    if noStokesI:
        if mpiRank==0:
            print("Note: no Stokes I data - assuming fractional polarisation.")
        IArr = np.ones_like(QArr)
        dIArr = np.zeros_like(QArr)

    # Convert to GHz for convenience
    freqArr_GHz = freqArr_Hz / 1e9
    lamSqArr_m2 = np.power(C/freqArr_Hz, 2.0)

    # Fit the Stokes I spectrum and create the fractional spectra
    if mpiRank==0:
        dataArr = create_frac_spectra(freqArr=freqArr_GHz,
                                      IArr=IArr,
                                      QArr=QArr,
                                      UArr=UArr,
                                      dIArr=dIArr,
                                      dQArr=dQArr,
                                      dUArr=dUArr,
                                      polyOrd=polyOrd,
                                      verbose=True)
    else:
        dataArr = None
    if mpiSwitch:
        dataArr = mpiComm.bcast(dataArr, root=0)
    (IModArr, qArr, uArr, dqArr, duArr, IfitDict) = dataArr

    # Plot the data and the Stokes I model fit
    if mpiRank==0:
        print("Plotting the input data and spectral index fit.")
        freqHirArr_Hz =  np.linspace(freqArr_Hz[0], freqArr_Hz[-1], 10000)
        IModHirArr = poly5(IfitDict["p"])(freqHirArr_Hz/1e9)
        specFig = plt.figure(figsize=(10, 6))
        plot_Ipqu_spectra_fig(freqArr_Hz     = freqArr_Hz,
                              IArr           = IArr,
                              qArr           = qArr,
                              uArr           = uArr,
                              dIArr          = dIArr,
                              dqArr          = dqArr,
                              duArr          = duArr,
                              freqHirArr_Hz  = freqHirArr_Hz,
                              IModArr        = IModHirArr,
                              fig            = specFig)

        # Use the custom navigation toolbar
        try:
            specFig.canvas.toolbar.pack_forget()
            CustomNavbar(specFig.canvas, specFig.canvas.toolbar.window)
        except Exception:
            pass

        # Display the figure
        if showPlots:
            specFig.canvas.draw()
            specFig.show()

    #-------------------------------------------------------------------------#

    # Load the model and parameters from the relevant file
    if mpiSwitch:
        mpiComm.Barrier()
    if mpiRank==0:
        print("\nLoading the model from 'models_ns/m%d.py' ..."  % modelNum)
    #First check the working directory for a model. Failing that, try the install directory.
    try:
        spec=importlib.util.spec_from_file_location("m%d" % modelNum,"models_ns/m%d.py" % modelNum)
        mod = importlib.util.module_from_spec(spec)
        sys.modules[mod] = mod
        spec.loader.exec_module(mod)
    except FileNotFoundError:
        try:
            RMtools_dir=os.path.dirname(importlib.util.find_spec('RMtools_1D').origin)
            spec=importlib.util.spec_from_file_location("m%d" % modelNum,RMtools_dir+"/models_ns/m%d.py" % modelNum)
            mod = importlib.util.module_from_spec(spec)
            sys.modules[mod] = mod
            spec.loader.exec_module(mod)
        except:
            print('Model could not be found! Please make sure model is present either in {}/models_ns/, or in {}/RMtools_1D/models_ns/'.format(os.getcwd(),RMtools_dir))
            sys.exit()

    global model
    model = mod.model


    # Let's time the sampler
    if mpiRank==0:
        startTime = time.time()

    # Unpack the inParms structure
    parNames = [x["parname"] for x in mod.inParms]
    labels = [x["label"] for x in mod.inParms]
    values = [x["value"] for x in mod.inParms]
    bounds = [x["bounds"] for x in mod.inParms]
    priorTypes = [x["priortype"] for x in mod.inParms]
    wraps = [x["wrap"] for x in mod.inParms]
    nDim = len(priorTypes)
    fixedMsk = [0 if x=="fixed" else 1 for x in priorTypes]
    nFree = sum(fixedMsk)

    # Set the prior function given the bounds of each parameter
    prior = prior_call(priorTypes, bounds, values)

    # Set the likelihood function given the data
    lnlike = lnlike_call(parNames, lamSqArr_m2, qArr, dqArr, uArr, duArr)

    # Let's time the sampler
    if mpiRank==0:
        startTime = time.time()

    # Run nested sampling using PyMultiNest
    nestArgsDict = merge_two_dicts(init_mnest(), mod.nestArgsDict)
    nestArgsDict["n_params"]             = nDim
    nestArgsDict["n_dims"]               = nDim
    nestArgsDict["outputfiles_basename"] = nestOut
    nestArgsDict["LogLikelihood"]        = lnlike
    nestArgsDict["Prior"]                = prior
    # Look for multiple modes
    nestArgsDict['multimodal']           = True
    nestArgsDict['n_clustering_params']  = nDim
    nestArgsDict['verbose']              = False
    pmn.run(**nestArgsDict)
    # Save parnames for use with PyMultinest tools
    json.dump(parNames, open(f'{nestOut}/params.json', 'w'))

    # Do the post-processing on one processor
    if mpiSwitch:
        mpiComm.Barrier()
    if mpiRank==0:

        # Query the analyser object for results
        aObj = pmn.Analyzer(n_params=nDim, outputfiles_basename=nestOut)
        statDict = aObj.get_stats()
        fitDict = aObj.get_best_fit()
        endTime = time.time()

        # NOTE: The Analyser methods do not work well for parameters with
        # posteriors that overlap the wrap value. Use np.percentile instead.
        pMed = [None]*nDim
        for i in range(nDim):
            pMed[i] = statDict["marginals"][i]['median']
        lnLike = fitDict["log_likelihood"]
        lnEvidence = statDict["nested sampling global log-evidence"]
        dLnEvidence = statDict["nested sampling global log-evidence error"]

        # Get the best-fitting values & uncertainties directly from chains
        chains =  aObj.get_equal_weighted_posterior()
        chains = wrap_chains(chains, wraps, bounds, pMed)
        # Find the mode with the highest evidence
        mode_evidence = [mode['strictly local log-evidence'] for mode in statDict['modes']]
        # Get the max and std for modal value
        modes = np.array(statDict['modes'][np.argmax(mode_evidence)]['maximum a posterior'])
        sigmas = np.array(statDict['modes'][np.argmax(mode_evidence)]['sigma'])
        upper = modes + sigma_clip * sigmas
        lower = modes - sigma_clip * sigmas

        p = [None]*nDim
        errPlus = [None]*nDim
        errMinus = [None]*nDim
        g = lambda v: (v[1], v[2]-v[1], v[1]-v[0])
        for i in range(nDim):
            # Get stats around modal value
            idx = (chains[:,i] > lower[i]) & (chains[:,i] < upper[i])
            p[i], errPlus[i], errMinus[i] = \
                        g(np.percentile(chains[idx, i], [15.72, 50, 84.27]))

        # Calculate goodness-of-fit parameters
        nData = 2.0 * len(lamSqArr_m2)
        dof = nData - nFree -1
        chiSq = chisq_model(parNames, p, lamSqArr_m2, qArr, dqArr, uArr, duArr)
        chiSqRed = chiSq/dof
        AIC = 2.0*nFree - 2.0 * lnLike
        AICc = 2.0*nFree*(nFree+1)/(nData-nFree-1) - 2.0 * lnLike
        BIC = nFree * np.log(nData) - 2.0 * lnLike
        
        # Summary of run
        print("")
        print("-"*80)
        print("SUMMARY OF SAMPLING RUN:")
        print("#-PROCESSORS  = %d" % mpiSize)
        print("RUN-TIME      = %.2f" % (endTime-startTime))
        print("DOF           = %d" % dof)
        print("CHISQ:        = %.3g" % chiSq)
        print("CHISQ RED     = %.3g" % chiSqRed)
        print("AIC:          = %.3g" % AIC)
        print("AICc          = %.3g" % AICc)
        print("BIC           = %.3g" % BIC)
        print("ln(EVIDENCE)  = %.3g" % lnEvidence)
        print("dLn(EVIDENCE) = %.3g" % dLnEvidence)
        print("")
        print("-"*80)
        print("RESULTS:\n")
        for i in range(len(p)):
            print("%s = %.4g (+%3g, -%3g)" % \
                  (parNames[i], p[i], errPlus[i], errMinus[i]))
        print("-"*80)
        print("")
        
        # Create a save dictionary and store final p in values
        outFile = prefixOut + "_m%d_nest.json" % modelNum
        IfitDict["p"] = toscalar(IfitDict["p"].tolist())
        saveDict = {"parNames":      toscalar(parNames),
                    "labels":        toscalar(labels),
                    "values":        toscalar(p),
                    "errPlus":       toscalar(errPlus),
                    "errMinus":      toscalar(errMinus),
                    "bounds":        toscalar(bounds),
                    "priorTypes":    toscalar(priorTypes),
                    "wraps":         toscalar(wraps),
                    "dof":           toscalar(dof),
                    "chiSq":         toscalar(chiSq),
                    "chiSqRed":      toscalar(chiSqRed),
                    "AIC":           toscalar(AIC),
                    "AICc":          toscalar(AICc),
                    "BIC":           toscalar(BIC),
                    "ln(EVIDENCE) ": toscalar(lnEvidence),
                    "dLn(EVIDENCE)": toscalar(dLnEvidence),
                    "nFree":         toscalar(nFree),
                    "IfitDict":      IfitDict}
        json.dump(saveDict, open(outFile, "w"))
        print("Results saved in JSON format to:\n '%s'\n" % outFile)

        # Plot the posterior samples in a corner plot
        chains =  aObj.get_equal_weighted_posterior()
        chains = wrap_chains(chains, wraps, bounds, p)[:, :nDim]
        iFixed = [i for i, e in enumerate(fixedMsk) if e==0]
        chains = np.delete(chains, iFixed, 1)
        for i in sorted(iFixed, reverse=True):
            del(labels[i])
            del(p[i])

        cornerFig = corner.corner(xs      = chains,
                                  labels  = labels,
                                  range   = [(l,u) for l,u in zip(upper,lower)],
                                  truths  = p,
                                  quantiles = [0.1572, 0.8427],
                                  bins    = 30)

        # Save the posterior chains to ASCII file
        if verbose: print("Saving the posterior chains to ASCII file.")
        outFile = prefixOut + "_m%d_posteriorChains.dat" % modelNum
        if verbose: print("> %s" % outFile)
        np.savetxt(outFile, chains)

        # Plot the data and best-fitting model
        lamSqHirArr_m2 =  np.linspace(lamSqArr_m2[0], lamSqArr_m2[-1], 10000)
        freqHirArr_Hz = C / np.sqrt(lamSqHirArr_m2)
        IModArr = poly5(IfitDict["p"])(freqHirArr_Hz/1e9)
        pDict = {k:v for k, v in zip(parNames, p)}
        quModArr = model(pDict, lamSqHirArr_m2)
        model_dict = {
            'chains': chains,
            'model': model,
            'parNames': parNames,
            'values': values
        }
        specFig.clf()
        plot_Ipqu_spectra_fig(freqArr_Hz     = freqArr_Hz,
                              IArr           = IArr,
                              qArr           = qArr,
                              uArr           = uArr,
                              dIArr          = dIArr,
                              dqArr          = dqArr,
                              duArr          = duArr,
                              freqHirArr_Hz  = freqHirArr_Hz,
                              IModArr        = IModArr,
                              qModArr        = quModArr.real,
                              uModArr        = quModArr.imag,
                              model_dict     = model_dict,
                              fig            = specFig)
        specFig.canvas.draw()

        # Save the figures
        outFile = prefixOut + "fig_m%d_specfit.pdf" % modelNum
        specFig.savefig(outFile)
        print("Plot of best-fitting model saved to:\n '%s'\n" % outFile)
        outFile = prefixOut + "fig_m%d_corner.pdf" % modelNum
        cornerFig.savefig(outFile)
        print("Plot of posterior samples saved to \n '%s'\n" % outFile)

        # Display the figures
        if showPlots:
            specFig.show()
            cornerFig.show()
            print("> Press <RETURN> to exit ...", end="")
            sys.stdout.flush()
            input()

        # Clean up
        plt.close(specFig)
        plt.close(cornerFig)

    # Clean up MPI environment
    if mpiSwitch:
        MPI.Finalize()


#-----------------------------------------------------------------------------#
def prior_call(priorTypes, bounds, values):
    """Returns a function to transform (0-1) range to the distribution of
    values for each parameter. Note that a numpy vectorised version of this
    function fails because of type-errors."""

    def prior(p, nDim, nParms):
        for i in range(nDim):
            if priorTypes[i] == "log":
                bMin = np.log(np.abs(bounds[i][0]))
                bMax = np.log(np.abs(bounds[i][1]))
                p[i] *= bMax - bMin
                p[i] += bMin
                p[i] = np.exp(p[i])
            elif priorTypes[i] == "normal":
                bMin, bMax = bounds[i]
                sigma = (bMax - bMin)/2.0
                mu = bMin + sigma
                p[i] = mu + sigma * ndtri(p[i])
            elif priorTypes[i] == "fixed":
                p[i] = values[i]
            else: # uniform (linear)
                bMin, bMax = bounds[i]
                p[i] = bMin + p[i] * (bMax - bMin)
        return p

    return prior


#-----------------------------------------------------------------------------#
def lnlike_call(parNames, lamSqArr_m2, qArr, dqArr, uArr, duArr):
    """ Returns a function to evaluate the log-likelihood """

    def lnlike(p, nDim, nParms):

        # Evaluate the model and calculate the joint ln(like)
        # Silva 2006
        pDict = {k:v for k, v in zip(parNames, p)}
        quMod = model(pDict, lamSqArr_m2)
        dquArr = np.sqrt(np.power(dqArr, 2) + np.power(duArr, 2))
        chiSqQ = np.nansum(np.power((qArr-quMod.real)/dqArr, 2))
        chiSqU = np.nansum(np.power((uArr-quMod.imag)/dqArr, 2))
        nData = len(dquArr)
        logLike = (-nData * np.log(2.0*np.pi)
                   -2.0 * np.nansum(np.log(dquArr))
                   -chiSqQ/2.0 -chiSqU/2.0)

        return logLike

    return lnlike


#-----------------------------------------------------------------------------#
def chisq_model(parNames, p, lamSqArr_m2, qArr, dqArr, uArr, duArr):
    """Calculate the chi^2 for the current model, given the data."""

    # Evaluate the model and calculate chisq
    pDict = {k:v for k, v in zip(parNames, p)}
    quMod = model(pDict, lamSqArr_m2)
    chisq = np.nansum( np.power((qArr-quMod.real)/dqArr, 2) +
                       np.power((uArr-quMod.imag)/duArr, 2))

    return chisq


#-----------------------------------------------------------------------------#
def wrap_chains(chains, wraps, bounds, p, verbose=False):

    # Get the indices of the periodic parameters
    iWrap = [i for i, e in enumerate(wraps) if e != 0]

    # Loop through the chains for periodic parameters
    for i in iWrap:
        wrapLow = bounds[i][0]
        wrapHigh = bounds[i][1]
        rng = wrapHigh - wrapLow

        # Shift the wrapping to centre on the best fit value
        wrapCent = wrapLow + (wrapHigh - wrapLow)/2.0
        wrapLow += (p[i] - wrapCent)
        wrapHigh += (p[i] - wrapCent)
        chains[:, i] = ((chains[:, i]-wrapLow) % rng) + wrapLow
        if verbose:
            print("Wrapped parameter '%d' in range [%s, %s] ..." %
                  (i, wrapLow, wrapHigh))

    return chains


#-----------------------------------------------------------------------------#
def init_mnest():
    """Initialise MultiNest arguments"""

    argsDict = {'LogLikelihood':              '',
                'Prior':                      '',
                'n_dims':                     0,
                'n_params':                   0,
                'n_clustering_params':        0,
                'wrapped_params':             None,
                'importance_nested_sampling': False,
                'multimodal':                 False,
                'const_efficiency_mode':      False,
                'n_live_points':              100,
                'evidence_tolerance':         0.5,
                'sampling_efficiency':        'model',
                'n_iter_before_update':       500,
                'null_log_evidence':          -1.e90,
                'max_modes':                  100,
                'mode_tolerance':             -1.e90,
                'outputfiles_basename':       '',
                'seed':                       -1,
                'verbose':                    True,
                'resume':                     True,
                'context':                    0,
                'write_output':               True,
                'log_zero':                   -1.e100,
                'max_iter':                   0,
                'init_MPI':                   False,
                'dump_callback':              None}
    return argsDict


#-----------------------------------------------------------------------------#
def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z

#-----------------------------------------------------------------------------#
if __name__ == "__main__":
    main()



#!/usr/bin/env python
#=============================================================================#
#                                                                             #
# NAME:     do_RM-clean.py                                                    #
#                                                                             #
# PURPOSE:  Run RM-clean on a  cube of dirty Faraday dispersion functions.    #
#                                                                             #
# MODIFIED: 15-May-2016 by C. Purcell                                         #
# MODIFIED: 23-October-2019 by A. Thomson                                     #
#                                                                             #
#=============================================================================#
#                                                                             #
# The MIT License (MIT)                                                       #
#                                                                             #
# Copyright (c) 2015 Cormac R. Purcell                                        #
#                                                                             #
# Permission is hereby granted, free of charge, to any person obtaining a     #
# copy of this software and associated documentation files (the "Software"),  #
# to deal in the Software without restriction, including without limitation   #
# the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# and/or sell copies of the Software, and to permit persons to whom the       #
# Software is furnished to do so, subject to the following conditions:        #
#                                                                             #
# The above copyright notice and this permission notice shall be included in  #
# all copies or substantial portions of the Software.                         #
#                                                                             #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# DEALINGS IN THE SOFTWARE.                                                   #
#                                                                             #
#=============================================================================#

import sys
import os
import time
import numpy as np
import astropy.io.fits as pf

try:
    import schwimmbad
    parallel_available=True
except:
    parallel_available=False

from RMutils.util_RM import do_rmclean_hogbom
from RMutils.util_RM import fits_make_lin_axis

C = 2.997924538e8 # Speed of light [m/s]


#-----------------------------------------------------------------------------#
def run_rmclean(fitsFDF, fitsRMSF, cutoff, maxIter=1000, gain=0.1, nBits=32,
                pool=None, chunksize=None, verbose = True, log = print):
    """Run RM-CLEAN on a 2/3D FDF cube given an RMSF cube stored as FITS.

    If you want to run RM-CLEAN on arrays, just use util_RM.do_rmclean_hogbom.

    Args:
        fitsFDF (str): Name of FDF FITS file.
        fitsRMSF (str): Name of RMSF FITS file
        cutoff (float): CLEAN cutoff in flux units

    Kwargs:
        maxIter (int): Maximum number of CLEAN iterations per pixel.
        gain (float): CLEAN loop gain.
        nBits (int): Precision of floating point numbers.
        pool (multiprocessing Pool): Pool function from multiprocessing
            or schwimmbad
        chunksize (int): Number of chunks which it submits to the process pool
            as separate tasks. The (approximate) size of these chunks can be
            specified by setting chunksize to a positive integer.
        verbose (bool): Verbosity.
        log (function): Which logging function to use.

    Returns:
        cleanFDF (ndarray): Cube of RMCLEANed FDFs.
        ccArr (ndarray): Cube of RMCLEAN components (i.e. the model).
        iterCountArr (ndarray): Cube of number of RMCLEAN iterations.
        residFDF (ndarray): Cube of residual RMCLEANed FDFs.
        head (fits.header): Header of FDF FITS file for template.

    """


    # Default data types
    dtFloat = "float" + str(nBits)
    dtComplex = "complex" + str(2*nBits)

    # Read the FDF
    dirtyFDF, head, FD_axis = read_FDF_cubes(fitsFDF)



    phiArr_radm2 = fits_make_lin_axis(head, axis=FD_axis-1, dtype=dtFloat)

    # Read the RMSF

    RMSFArr, headRMSF,FD_axis = read_FDF_cubes(fitsRMSF)
    HDULst = pf.open(fitsRMSF.replace('_real','_FWHM').replace('_im','_FWHM').replace('_tot','_FWHM'), "readonly", memmap=True)
    fwhmRMSFArr = np.squeeze(HDULst[0].data)
    HDULst.close()
    phi2Arr_radm2 = fits_make_lin_axis(headRMSF, axis=FD_axis-1, dtype=dtFloat)

    startTime = time.time()

    # Do the clean
    cleanFDF, ccArr, iterCountArr, residFDF  = \
        do_rmclean_hogbom(dirtyFDF         = dirtyFDF,
                          phiArr_radm2     = phiArr_radm2,
                          RMSFArr          = RMSFArr,
                          phi2Arr_radm2    = phi2Arr_radm2,
                          fwhmRMSFArr      = fwhmRMSFArr,
                          cutoff           = cutoff,
                          maxIter          = maxIter,
                          gain             = gain,
                          verbose          = verbose,
                          doPlots          = False,
                          pool             = pool,
                          chunksize        = chunksize)


    endTime = time.time()
    cputime = (endTime - startTime)
    if (verbose): log("> RM-clean completed in %.2f seconds." % cputime)
    if (verbose): log("Saving the clean FDF and ancillary FITS files")




    #Move FD axis back to original position, and restore dimensionality:
    old_Ndim,FD_axis=find_axes(head)
    new_Ndim=cleanFDF.ndim
    #The difference is the number of dimensions that need to be added:
    if old_Ndim-new_Ndim != 0: #add missing (degenerate) dimensions back in
        cleanFDF=np.expand_dims(cleanFDF,axis=tuple(range(old_Ndim-new_Ndim)))
        ccArr=np.expand_dims(ccArr,axis=tuple(range(old_Ndim-new_Ndim)))
        residFDF=np.expand_dims(residFDF,axis=tuple(range(old_Ndim-new_Ndim)))
    #New dimensions are added to the beginning of the axis ordering 
    #(revserse of FITS ordering)
    
    #Move the FDF axis to it's original spot. Hopefully this means that all
    # axes are in their original place after all of that.
    cleanFDF=np.moveaxis(cleanFDF,old_Ndim-new_Ndim,old_Ndim-FD_axis)
    ccArr=np.moveaxis(ccArr,old_Ndim-new_Ndim,old_Ndim-FD_axis)
    residFDF=np.moveaxis(residFDF,old_Ndim-new_Ndim,old_Ndim-FD_axis)

    return cleanFDF, ccArr, iterCountArr, residFDF, head


def writefits(cleanFDF, ccArr, iterCountArr, residFDF, headtemp, nBits=32,
            prefixOut="", outDir="", write_separate_FDF=False, verbose=True, log=print):
    """Write data to disk in FITS


    Output files:
        Default:
            FDF_clean.fits: RMCLEANed FDF, in 3 extensions: Q,U, and PI.
            FDF_CC.fits: RMCLEAN components, in 3 extensions: Q,U, and PI.
            CLEAN_nIter.fits: RMCLEAN iterations.

        write_seperate_FDF=True:
            FDF_clean_real.fits and FDF_CC.fits are split into
            three constituent components:
                FDF_clean_real.fits: Stokes Q
                FDF_clean_im.fits: Stokes U
                FDF_clean_tot.fits: Polarized Intensity (sqrt(Q^2+U^2))
                FDF_CC_real.fits: Stokes Q
                FDF_CC_im.fits: Stokes U
                FDF_CC_tot.fits: Polarized Intensity (sqrt(Q^2+U^2))
                CLEAN_nIter.fits: RMCLEAN iterations.
    Args:
        cleanFDF (ndarray): Cube of RMCLEANed FDFs.
        ccArr (ndarray): Cube of RMCLEAN components (i.e. the model).
        iterCountArr (ndarray): Cube of number of RMCLEAN iterations.
        residFDF (ndarray): Cube of residual RMCLEANed FDFs.

    Kwargs:
        prefixOut (str): Prefix for filenames.
        outDir (str): Directory to save files.
        write_seperate_FDF (bool): Write Q, U, and PI separately?
        verbose (bool): Verbosity.
        log (function): Which logging function to use.
    """
    # Default data types
    dtFloat = "float" + str(nBits)
    dtComplex = "complex" + str(2*nBits)


    if outDir=='':  #To prevent code breaking if file is in current directory
        outDir='.'
    # Save the clean FDF
    if not write_separate_FDF:
        fitsFileOut = outDir + "/" + prefixOut + "FDF_clean.fits"
        if(verbose): log("> %s" % fitsFileOut)
        hdu0 = pf.PrimaryHDU(cleanFDF.real.astype(dtFloat), headtemp)
        hdu1 = pf.ImageHDU(cleanFDF.imag.astype(dtFloat), headtemp)
        hdu2 = pf.ImageHDU(np.abs(cleanFDF).astype(dtFloat), headtemp)
        hduLst = pf.HDUList([hdu0, hdu1, hdu2])
        hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        hduLst.close()
    else:
        hdu0 = pf.PrimaryHDU(cleanFDF.real.astype(dtFloat), headtemp)
        fitsFileOut = outDir + "/" + prefixOut + "FDF_clean_real.fits"
        hdu0.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        if (verbose): log("> %s" % fitsFileOut)
        hdu1 = pf.PrimaryHDU(cleanFDF.imag.astype(dtFloat), headtemp)
        fitsFileOut = outDir + "/" + prefixOut + "FDF_clean_im.fits"
        hdu1.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        if (verbose): log("> %s" % fitsFileOut)
        hdu2 = pf.PrimaryHDU(np.abs(cleanFDF).astype(dtFloat), headtemp)
        fitsFileOut = outDir + "/" + prefixOut + "FDF_clean_tot.fits"
        hdu2.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        if (verbose): log("> %s" % fitsFileOut)

    if not write_separate_FDF:
    #Save the complex clean components as another file.
        fitsFileOut = outDir + "/" + prefixOut + "FDF_CC.fits"
        if (verbose): log("> %s" % fitsFileOut)
        hdu0 = pf.PrimaryHDU(ccArr.real.astype(dtFloat), headtemp)
        hdu1 = pf.ImageHDU(ccArr.imag.astype(dtFloat), headtemp)
        hdu2 = pf.ImageHDU(np.abs(ccArr).astype(dtFloat), headtemp)
        hduLst = pf.HDUList([hdu0, hdu1, hdu2])
        hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        hduLst.close()
    else:
        hdu0 = pf.PrimaryHDU(ccArr.real.astype(dtFloat), headtemp)
        fitsFileOut = outDir + "/" + prefixOut + "FDF_CC_real.fits"
        hdu0.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        if (verbose): log("> %s" % fitsFileOut)
        hdu1 = pf.PrimaryHDU(ccArr.imag.astype(dtFloat), headtemp)
        fitsFileOut = outDir + "/" + prefixOut + "FDF_CC_im.fits"
        hdu1.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        if (verbose): log("> %s" % fitsFileOut)
        hdu2 = pf.PrimaryHDU(np.abs(ccArr).astype(dtFloat), headtemp)
        fitsFileOut = outDir + "/" + prefixOut + "FDF_CC_tot.fits"
        hdu2.writeto(fitsFileOut, output_verify="fix", overwrite=True)
        if (verbose): log("> %s" % fitsFileOut)

    #Because there can be problems with different axes having different FITS keywords,
    #don't try to remove the FD axis, but just make it degenerate.

    if headtemp['NAXIS'] > 2:
        headtemp["NAXIS3"] = 1
    if headtemp['NAXIS'] == 4:
        headtemp["NAXIS4"] = 1

    # Save the iteration count mask
    fitsFileOut = outDir + "/" + prefixOut + "CLEAN_nIter.fits"
    if (verbose): log("> %s" % fitsFileOut)
    headtemp["BUNIT"] = "Iterations"
    hdu0 = pf.PrimaryHDU(np.expand_dims(iterCountArr.astype(dtFloat), 
                        axis=tuple(range(headtemp['NAXIS']-iterCountArr.ndim))),
                        headtemp)
    hduLst = pf.HDUList([hdu0])
    hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
    hduLst.close()


#Old method (for multi-extension files)
def read_FDF_cube(filename):
    """Read in a FDF/RMSF cube. Figures out which axis is Faraday depth and
    puts it first (in numpy order) to accommodate the rest of the code.
    Returns: (complex_cube, header,FD_axis)
    """
    HDULst = pf.open(filename, "readonly", memmap=True)
    head = HDULst[0].header.copy()
    FDFreal = HDULst[0].data
    FDFimag = HDULst[1].data
    complex_cube = FDFreal + 1j * FDFimag

    #Identify Faraday depth axis (assumed to be last one if not explicitly found)
    Ndim=head['NAXIS']
    FD_axis=Ndim
    #Check for FD axes:
    for i in range(1,Ndim+1):
        try:
            if 'FDEP' in head['CTYPE'+str(i)].upper():
                FD_axis=i
        except:
            pass #The try statement is needed for if the FITS header does not
                 # have CTYPE keywords.

    #Move FD axis to first place in numpy order.
    if FD_axis != Ndim:
        complex_cube=np.moveaxis(complex_cube,Ndim-FD_axis,0)

    #Remove degenerate axes to prevent problems with later steps.
    complex_cube=complex_cube.squeeze()

    return complex_cube, head,FD_axis

def find_axes(header):
    """Idenfities how many axes are present in a FITS file, and which is the
    Faraday depth axis. Necessary for bookkeeping on cube dimensionality,
    given that RM-clean only supports 3D cubes, but data may be 4D files."""
    Ndim=header['NAXIS']
    FD_axis=Ndim
    #Check for FD axes:
    for i in range(1,Ndim+1):
        try:
            if 'FDEP' in header['CTYPE'+str(i)].upper():
                FD_axis=i
        except:
            pass #The try statement is needed for if the FITS header does not
                 # have CTYPE keywords.
    return Ndim,FD_axis

def read_FDF_cubes(filename):
    """Read in a FDF/RMSF cube. Input filename can be any of real, imag, or tot components.
    Figures out which axis is Faraday depth and
    puts it first (in numpy order) to accommodate the rest of the code.
    Returns: (complex_cube, header,FD_axis)
    """
    HDUreal = pf.open(filename.replace('_tot','_real').replace('_im','_real'), "readonly", memmap=True)
    head = HDUreal[0].header.copy()
    FDFreal = HDUreal[0].data

    HDUimag = pf.open(filename.replace('_tot','_im').replace('_real','_im'), "readonly", memmap=True)
    FDFimag = HDUimag[0].data
    complex_cube = FDFreal + 1j * FDFimag

    #Identify Faraday depth axis (assumed to be last one if not explicitly found)
    Ndim,FD_axis=find_axes(head)

    #Move FD axis to first place in numpy order.
    if FD_axis != Ndim:
        complex_cube=np.moveaxis(complex_cube,Ndim-FD_axis,0)

    #Remove degenerate axes to prevent problems with later steps.
    complex_cube=complex_cube.squeeze()

    return complex_cube, head,FD_axis


#-----------------------------------------------------------------------------#
def main():
    import argparse
    """
    Start the function to perform RM-clean if called from the command line.
    """

    # Help string to be shown using the -h option
    descStr = """
    Run RM-CLEAN on a cube of Faraday dispersion functions (FDFs), applying
    a cube of rotation measure spread functions created by the script
    'do_RMsynth_3D.py'. Saves cubes of deconvolved FDFs & clean-component
    spectra, and a pixel map showing the number of iterations performed.
    Set any of the multiprocessing options to enable parallelization
    (otherwise, pixels will be processed serially).

    Expects that the input is in the form of the Stokes-separated
    (single extension) FITS cubes produced by do_RMsynth_3D.
    """

    # Parse the command line options
    parser = argparse.ArgumentParser(description=descStr,
                                 formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("fitsFDF", metavar="FDF_dirty.fits", nargs=1,
                        help="FITS cube containing the dirty FDF.\n(Can be any of the FDF output cubes from do_RMsynth_3D.py)")
    parser.add_argument("fitsRMSF", metavar="RMSF.fits", nargs=1,
                        help="FITS cube containing the RMSF and FWHM image.\n(Cans be any of the RMSF output cubes from do_RMsynth_3D.py)")
    parser.add_argument("-c", dest="cutoff", type=float, nargs=1,
                        default=1.0, help="CLEAN cutoff in flux units")
    parser.add_argument("-n", dest="maxIter", type=int, default=1000,
                        help="Maximum number of CLEAN iterations per pixel [1000].")
    parser.add_argument("-g", dest="gain", type=float, default=0.1,
                        help="CLEAN loop gain [0.1].")
    parser.add_argument("-o", dest="prefixOut", default="",
                        help="Prefix to prepend to output files [None].")
    parser.add_argument("-f", dest="write_separate_FDF", action="store_false",
                        help="Store different Stokes as FITS extensions [False, store as separate files].")

    parser.add_argument("-v", dest="verbose", action="store_true",
                        help="Verbose [False].")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("--ncores", dest="n_cores", default=1,
                       type=int, help="Number of processes (uses multiprocessing).")
    parser.add_argument("--chunk", dest="chunk", default=None,
                       type=int, help="Chunk size (uses multiprocessing -- not available in MPI!)")
    group.add_argument("--mpi", dest="mpi", default=False,
                       action="store_true", help="Run with MPI.")


    args = parser.parse_args()

    #Check if the user is trying to use parallelization without installing it:
    if parallel_available==False and (args.n_cores != 1 or args.chunk != None or args.mpi != False):
        raise Exception('Parallel processing not available. Please install the schwimmbad module to enable parallel processing.')

    #If parallelization requested use it, otherwise use the old-fashioned way.
    if parallel_available==True and (args.n_cores != 1 or args.chunk != None or args.mpi != False):
        pool = schwimmbad.choose_pool(mpi=args.mpi, processes=args.n_cores)
        if args.mpi:
            if not pool.is_master():
                pool.wait()
                sys.exit(0)
        if args.n_cores > 1:
            chunksize = args.chunk
        else:
            chunksize = None
    else:
        pool = None
        chunksize = None


    verbose = args.verbose
    # Sanity checks
    for f in args.fitsFDF + args.fitsRMSF:
        if not os.path.exists(f):
            print("File does not exist: '%s'." % f)
            sys.exit()
    dataDir, dummy = os.path.split(args.fitsFDF[0])

    # Run RM-CLEAN on the cubes
    cleanFDF, ccArr, iterCountArr, residFDF, headtemp = run_rmclean(fitsFDF     = args.fitsFDF[0],
                                                        fitsRMSF    = args.fitsRMSF[0],
                                                        cutoff      = args.cutoff,
                                                        maxIter     = args.maxIter,
                                                        gain        = args.gain,
                                                        chunksize   = chunksize,
                                                        nBits       = 32,
                                                        verbose = verbose)
    # Write results to disk
    writefits(cleanFDF,
            ccArr,
            iterCountArr,
            residFDF,
            headtemp,
            prefixOut           = args.prefixOut,
            outDir              = dataDir,
            write_separate_FDF  = args.write_separate_FDF,
            verbose             = verbose)

#-----------------------------------------------------------------------------#
if __name__ == "__main__":
    main()



# License is MIT: see LICENSE.md.
"""Nestle: nested sampling routines to evaluate Bayesian evidence."""



import sys
import warnings
import math

import numpy as np
try:
    from scipy.cluster.vq import kmeans2
    HAVE_KMEANS = True
except ImportError:  # pragma: no cover
    HAVE_KMEANS = False

__all__ = ["sample", "print_progress", "mean_and_cov", "resample_equal",
           "Result"]
__version__ = "0.2.0"

SQRTEPS = math.sqrt(float(np.finfo(np.float64).eps))

# -----------------------------------------------------------------------------
# Helpers

def vol_prefactor(n):
    """Volume constant for an n-dimensional sphere:

    for n even:      (2pi)^(n    /2) / (2 * 4 * ... * n)
    for n odd :  2 * (2pi)^((n-1)/2) / (1 * 3 * ... * n)
    """
    if n % 2 == 0:
        f = 1.
        i = 2
        while i <= n:
            f *= (2. / i * math.pi)
            i += 2
    else:
        f = 2.
        i = 3
        while i <= n:
            f *= (2. / i * math.pi)
            i += 2

    return f


def randsphere(n, rstate=np.random):
    """Draw a random point within an n-dimensional unit sphere"""

    z = rstate.randn(n)
    return z * rstate.rand()**(1./n) / np.sqrt(np.sum(z**2))


def random_choice(a, p, rstate=np.random):
    """replacement for numpy.random.choice (only in numpy 1.7+)"""

    if abs(np.sum(p) - 1.) > SQRTEPS:  # same tol as in np.random.choice.
        raise ValueError("probabilities do not sum to 1")

    r = rstate.rand()
    i = 0
    t = p[i]
    while t < r:
        i += 1
        t += p[i]
    return i


def resample_equal(samples, weights, rstate=None):
    """Resample the samples so that the final samples all have equal weight.

    Each input sample appears in the output array either
    `floor(weights[i] * N)` or `ceil(weights[i] * N)` times, with
    `floor` or `ceil` randomly selected (weighted by proximity).

    Parameters
    ----------
    samples : `~numpy.ndarray`
        Unequally weight samples returned by the nested sampling algorithm.
        Shape is (N, ...), with N the number of samples.

    weights : `~numpy.ndarray`
        Weight of each sample. Shape is (N,).
    
    Returns
    -------
    equal_weight_samples : `~numpy.ndarray`
        Samples with equal weights, same shape as input samples.

    Examples
    --------

    >>> x = np.array([[1., 1.], [2., 2.], [3., 3.], [4., 4.]])
    >>> w = np.array([0.6, 0.2, 0.15, 0.05])
    >>> nestle.resample_equal(x, w)
    array([[ 1.,  1.],
           [ 1.,  1.],
           [ 1.,  1.],
           [ 3.,  3.]])

    Notes
    -----
    Implements the systematic resampling method described in
    `this PDF <http://people.isy.liu.se/rt/schon/Publications/HolSG2006.pdf>`_.
    Another way to sample according to weights would be::

        N = len(weights)
        new_samples = samples[np.random.choice(N, size=N, p=weights)]

    However, the method used in this function is less "noisy".

    """

    if abs(np.sum(weights) - 1.) > SQRTEPS:  # same tol as in np.random.choice.
        raise ValueError("weights do not sum to 1")

    if rstate is None:
        rstate = np.random

    N = len(weights)

    # make N subdivisions, and choose positions with a consistent random offset
    positions = (rstate.random() + np.arange(N)) / N

    idx = np.zeros(N, dtype=np.int)
    cumulative_sum = np.cumsum(weights)
    i, j = 0, 0
    while i < N:
        if positions[i] < cumulative_sum[j]:
            idx[i] = j
            i += 1
        else:
            j += 1

    return samples[idx]


class Result(dict):
    """Represents a sampling result.

    Since this class is essentially a subclass of dict with attribute
    accessors, one can see which attributes are available using the
    `keys()` method.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if list(self.keys()):
            m = max(list(map(len, list(self.keys())))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in list(self.items())])
        else:
            return self.__class__.__name__ + "()"

    def summary(self):
        """Return a nicely formatted string giving summary."""
        return ("niter: {:d}\n"
                "ncall: {:d}\n"
                "nsamples: {:d}\n"
                "logz: {:6.3f} +/- {:6.3f}\n"
                "h: {:6.3f}"
                .format(self.niter, self.ncall, len(self.samples),
                        self.logz, self.logzerr, self.h))


def mean_and_cov(x, weights):
    """Compute weighted sample mean and covariance.

    Parameters
    ----------
    x : `~numpy.ndarray`
        2-D array containing data samples. Shape is (M, N) where N is the
        number of variables and M is the number of samples or observations.
        This is ordering is equivalent to using ``rowvar=0`` in numpy.cov.
    weights : `~numpy.ndarray`
        1-D array of sample weights. Shape is (M,).

    Returns
    -------
    mean : `~numpy.ndarray`
        Weighted average of samples, with shape (N,).
    cov : `~numpy.ndarray`
        The covariance matrix of the variables with shape (N, N).

    Notes
    -----
    Implements formula described here:
    https://en.wikipedia.org/wiki/Sample_mean_and_sample_covariance
    (see "weighted samples" section)
    """

    mean = np.average(x, weights=weights, axis=0)
    dx = x - mean
    wsum = np.sum(weights)
    w2sum = np.sum(weights**2)

    cov = wsum / (wsum**2 - w2sum) * np.einsum('i,ij,ik', weights, dx, dx)

    return mean, cov


def print_progress(info):
    """Callback function that prints a running total on a single line.

    Parameters
    ----------
    info : dict
        Dictionary containing keys ``'it'`` and ``'logz'``.
    """

    print("\r\033[Kit={:6d} logz={:8f}".format(info['it'], info['logz']),
          end='')
    sys.stdout.flush()  # because flush keyword not in print() in py2.7


# -----------------------------------------------------------------------------
# Ellipsoid

class Ellipsoid(object):
    """An N-ellipsoid.

    Defined by::

        (x - v)^T A (x - v) = 1

    where the vector ``v`` is the center of the ellipse and ``A`` is an N x N
    matrix. Assumes that ``A`` is symmetric positive definite.

    Parameters
    ----------
    ctr : `~numpy.ndarray` with shape ``(N,)``
        Coordinates of ellipse center. Note that the array is *not* copied.
        This array is never modified internally.
    a : `~numpy.ndarray` with shape ``(N, N)``
        Matrix describing the axes. Watch out! This array is *not* copied.
        but may be modified internally!
    """

    def __init__(self, ctr, a):
        self.n = len(ctr)
        self.ctr = ctr    # center coordinates
        self.a = a        # ~ inverse of covariance of points contained
        self.vol = vol_prefactor(self.n) / np.sqrt(np.linalg.det(a))

        # eigenvalues (l) are a^-2, b^-2, ... (lengths of principle axes)
        # eigenvectors (v) are normalized principle axes
        l, v = np.linalg.eigh(a)
        self.axlens = 1. / np.sqrt(l)

        # Scaled eigenvectors are the axes: axes[:,i] is the i-th
        # axis.  Multiplying this matrix by a vector will transform a
        # point in the unit n-sphere into a point in the ellipsoid.
        self.axes = np.dot(v, np.diag(self.axlens))

    def scale_to_vol(self, vol):
        """Scale ellipoid to satisfy a target volume."""
        f = (vol / self.vol) ** (1.0 / self.n)  # linear factor
        self.a *= f**-2
        self.axlens *= f
        self.axes *= f
        self.vol = vol

    def major_axis_endpoints(self):
        """Return the endpoints of the major axis"""
        i = np.argmax(self.axlens)  # which is the major axis?
        v = self.axes[:, i]  # vector to end of major axis
        return self.ctr - v, self.ctr + v

    def contains(self, x):
        """Does the ellipse contain the point?"""
        d = x - self.ctr
        return np.dot(np.dot(d, self.a), d) <= 1.0

    def randoffset(self, rstate=np.random):
        """Return an offset from ellipsoid center, randomly distributed
        within ellipsoid."""
        return np.dot(self.axes, randsphere(self.n, rstate=rstate))

    def sample(self, rstate=np.random):
        """Chose a sample randomly distributed within the ellipsoid.

        Returns
        -------
        x : 1-d array
            A single point within the ellipsoid.
        """
        return self.ctr + self.randoffset(rstate=rstate)

    def samples(self, nsamples, rstate=np.random):
        """Chose a sample randomly distributed within the ellipsoid.

        Returns
        -------
        x : (nsamples, ndim) array
            Coordinates within the ellipsoid.
        """

        x = np.empty((nsamples, self.n), dtype=np.float)
        for i in range(nsamples):
            x[i, :] = self.sample(rstate=rstate)
        return x

    def __repr__(self):
        return "Ellipsoid(ctr={})".format(self.ctr)


# -----------------------------------------------------------------------------
# Functions for determining the ellipsoid or set of ellipsoids bounding a
# set of points.

def make_eigvals_positive(a, targetprod):
    """For the symmetric square matrix ``a``, increase any zero eigenvalues
    to fulfill the given target product of eigenvalues.

    Returns a (possibly) new matrix."""

    w, v = np.linalg.eigh(a)  # Use eigh because we assume a is symmetric.
    mask = w < 1.e-10
    if np.any(mask):
        nzprod = np.product(w[~mask])  # product of nonzero eigenvalues
        nzeros = mask.sum()  # number of zero eigenvalues
        w[mask] = (targetprod / nzprod) ** (1./nzeros)  # adjust zero eigvals
        a = np.dot(np.dot(v, np.diag(w)), np.linalg.inv(v))  # re-form cov

    return a


def bounding_ellipsoid(x, pointvol=0., minvol=False):
    """Calculate bounding ellipsoid containing a set of points x.

    Parameters
    ----------
    x : (npoints, ndim) ndarray
        Coordinates of points.
    pointvol : float, optional
        Used to set a minimum bound on the ellipsoid volume when
        minvol is True.
    minvol : bool, optional
        If True, ensure that ellipsoid volume is at least len(x) * pointvol.

    Returns
    -------
    ellipsoid : Ellipsoid
    """
    npoints, ndim = x.shape

    # If there is only a single point, return an N-sphere with volume `pointvol`
    # centered at the point.
    if npoints == 1:
        r = (pointvol / vol_prefactor(ndim))**(1./ndim)
        return Ellipsoid(x[0], (1. / r**2) * np.identity(ndim))

    # Calculate covariance of points
    ctr = np.mean(x, axis=0)
    delta = x - ctr
    cov = np.cov(delta, rowvar=0)
    
    # when ndim = 1, np.cov returns a 0-d array. Make it a 1x1 2-d array.
    if ndim == 1:
        cov = np.atleast_2d(cov)

    # For a ball of uniformly distributed points, the covariance will be
    # smaller than r^2 by a factor of 1/(n+2) [see, e.g.,
    # http://mathoverflow.net/questions/35276/
    # covariance-of-points-distributed-in-a-n-ball]. In nested sampling,
    # we are supposing the points are uniformly distributed within
    # an ellipse, so the same factor holds. Expand `cov`
    # to compensate for that when defining the ellipse matrix:
    cov *= (ndim + 2)

    # Ensure that ``cov`` is nonsingular.
    # It can be singular when the ellipsoid has zero volume, which happens
    # when npoints <= ndim or when enough points are linear combinations
    # of other points. (e.g., npoints = ndim+1 but one point is a linear
    # combination of others). When this happens, we expand the ellipse
    # in the zero dimensions to fulfill the volume expected from
    # ``pointvol``.
    targetprod = (npoints * pointvol / vol_prefactor(ndim))**2
    cov = make_eigvals_positive(cov, targetprod)

    # The matrix defining the ellipsoid.
    a = np.linalg.inv(cov)

    # Calculate expansion factor necessary to bound each point.
    # Points should obey x^T A x <= 1, so we calculate x^T A x for
    # each point and then scale A up or down to make the
    # "outermost" point obey x^T A x = 1.
    # 
    # fast way to compute delta[i] @ A @ delta[i] for all i.
    f = np.einsum('...i, ...i', np.tensordot(delta, a, axes=1), delta)
    fmax = np.max(f)

    # Due to round-off errors, we actually scale the ellipse so the outermost
    # point obeys x^T A x < 1 - (a bit), so that all the points will
    # *definitely* obey x^T A x < 1.
    one_minus_a_bit = 1. - SQRTEPS

    if fmax > one_minus_a_bit:
        a *= one_minus_a_bit / fmax

    ell = Ellipsoid(ctr, a)

    if minvol:
        v = len(x) * pointvol
        if ell.vol < v:
            ell.scale_to_vol(v)

    return ell


def _bounding_ellipsoids(x, ell, pointvol=0.):
    """Internal bounding ellipsoids method for when a bounding ellipsoid for
    the entire set has already been calculated.

    Parameters
    ----------
    x : (npoints, ndim) ndarray
        Coordinates of points.
    ell : Ellipsoid, optional
        If known, the bounding ellipsoid of the points `x`. If not supplied,
        it will be calculated. This option is used when the function calls
        itself recursively.
    pointvol : float, optional
        Volume represented by a single point. Used when number of points
        per ellipsoid is less than number of dimensions in order to make
        volume non-zero.

    Returns
    -------
    ells : list of Ellipsoid
        Ellipsoids.
    """

    npoints, ndim = x.shape

    # starting cluster centers for kmeans (k=2)
    p1, p2 = ell.major_axis_endpoints()  # returns two 1-d arrays
    start_ctrs = np.vstack((p1, p2)) # shape is (k, N) = (2, N)

    # Split points into two clusters using k-means clustering with k=2
    # centroid = (2, ndim) ; label = (npoints,)
    # [Each entry in `label` is 0 or 1, corresponding to cluster number]
    centroid, label = kmeans2(x, k=start_ctrs, iter=10, minit='matrix')

    # Get points in each cluster.
    xs = [x[label == k, :] for k in (0, 1)]  # points in each cluster

    # If either cluster has less than ndim+1 points, the bounding ellipsoid
    # will be ill-constrained, so we reject the split and simply return the
    # ellipsoid bounding all the points.
    if xs[0].shape[0] < 2 * ndim or xs[1].shape[0] < 2 * ndim:
        return [ell]

    # Bounding ellipsoid for each cluster, enlarging to minimum volume.
    ells = [bounding_ellipsoid(xi, pointvol=pointvol, minvol=True)
            for xi in xs]

    # If the total volume decreased by a significant amount,
    # then we will accept the split into subsets and try to perform the
    # algorithm on each subset.
    if ells[0].vol + ells[1].vol < 0.5 * ell.vol:
        return (_bounding_ellipsoids(xs[0], ells[0], pointvol=pointvol) +
                _bounding_ellipsoids(xs[1], ells[1], pointvol=pointvol))

    # Otherwise, see if the total ellipse volume is significantly greater
    # than expected. If it is, this indicates that there may be more than 2
    # clusters and we should try to subdivide further.
    if ell.vol > 2. * npoints * pointvol:
        out = (_bounding_ellipsoids(xs[0], ells[0], pointvol=pointvol) +
               _bounding_ellipsoids(xs[1], ells[1], pointvol=pointvol))

        # only accept split if volume decreased significantly
        if sum(e.vol for e in out) < 0.5 * ell.vol:
            return out

    # Otherwise, we are happy with the single bounding ellipse.
    return [ell]


def bounding_ellipsoids(x, pointvol=0.):
    """Calculate a set of ellipses that bound the points.

    Parameters
    ----------
    x : (npoints, ndim) ndarray
        Coordinates of points.
    pointvol : float, optional
        Volume represented by a single point. Used when number of points
        per ellipsoid is less than number of dimensions in order to make
        volume non-zero.

    Returns
    -------
    ells : list of Ellipsoid
        Ellipsoids.
    """

    # Calculate a single bounding ellipsoid for the points, and enlarge it
    # so that it has at least the minimum volume.
    ell = bounding_ellipsoid(x, pointvol=pointvol, minvol=True)

    return _bounding_ellipsoids(x, ell, pointvol=pointvol)


def sample_ellipsoids(ells, rstate=np.random):
    """Chose sample(s) randomly distributed within a set of
    (possibly overlapping) ellipsoids.
    
    Parameters
    ----------
    ells : list of Ellipsoid

    Returns
    -------
    x : 1-d ndarray
        Coordinates within the ellipsoids. 
    """

    nells = len(ells)

    if nells == 1:
        return ells[0].sample(rstate=rstate)

    # Select an ellipsoid at random, according to volumes
    vols = np.array([ell.vol for ell in ells])
    i = random_choice(nells, vols / vols.sum(), rstate=rstate)
    
    # Select a point from the ellipsoid
    x = ells[i].sample(rstate=rstate)

    # How many ellipsoids is the sample in?
    n = 1
    for j in range(nells):
        if j == i:
            continue
        n += ells[j].contains(x)

    # Only accept the point with probability 1/n
    # (If rejected, sample again).
    if n == 1 or rstate.rand() < 1.0 / n:
        return x
    else:
        return sample_ellipsoids(ells, rstate=rstate)

# -----------------------------------------------------------------------------
# Classes for dealing with non-parallel calls

class FakePool(object):
    """A fake Pool for serial function evaluations."""

    def __init__(self):
        pass

    def submit(self, fn, *args, **kwargs):
        return FakeFuture(fn, *args, **kwargs)

    def map(self, func, *iterables):
        return list(map(func, *iterables))

    def shutdown(self):
        pass
    
class FakeFuture(object):
    """A fake Future to mimic function calls."""

    def __init__(self, fn, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def result(self):
        return self.fn(*self.args, **self.kwargs)

    def cancel(self):
        return True

# -----------------------------------------------------------------------------
# Sampler classes

class Sampler:
    """A sampler simply selects a new point obeying the likelihood bound,
    given some existing set of points."""

    def __init__(self, loglikelihood, prior_transform, points, rstate,
                 options, queue_size, pool):
        self.loglikelihood = loglikelihood
        self.prior_transform = prior_transform
        self.points = points
        self.rstate = rstate
        self.set_options(options)
        self.queue_size = queue_size
        self.pool = pool
        self.queue = []
        self.submitted = 0
        self.cancelled = 0
        self.unused = 0
        self.used = 0

    def empty_queue(self):
        """Dump all operations on the queue."""

        while self.queue:
            x, v, f = self.queue.pop()
            if f.cancel():
                self.cancelled += 1
            else:
                self.unused += 1

    def fill_queue(self):
        """Fill up the queue with operations."""

        while len(self.queue)<self.queue_size:
            x = self.propose_point()
            v = self.prior_transform(x)
            self.queue.append((x, v, self.pool.submit(self.loglikelihood, v)))
            self.submitted += 1

    def get_point_value(self):
        """ Get evaluation sequentially from the queue. If we should
            update our proposal distribution, do not refill the queue."""

        if not self.queue:
            self.fill_queue()
        x, v, f = self.queue.pop(0)
        r = f.result()
        self.fill_queue()
        self.used += 1
        return x, v, r

class ClassicSampler(Sampler):
    """Picks an active point at random and evolves it with a
    Metropolis-Hastings style MCMC with fixed number of iterations."""

    def set_options(self, options):
        self.steps = options.get('steps', 20)

    def update(self, pointvol):
        """Calculate an ellipsoid to get the rough shape of the point
        distribution correct, but then scale it down to the volume
        corresponding to a single point."""

        self.ell = bounding_ellipsoid(self.points, pointvol=pointvol)
        self.ell.scale_to_vol(pointvol)

    def propose_point(self, u, scale):
        while True:
            new_u = u + scale * self.ell.randoffset(rstate=self.rstate)
            if np.all(new_u > 0.) and np.all(new_u < 1.):
                break
        return new_u

    def new_point(self, loglstar):
        # choose a point at random and copy it
        i = self.rstate.randint(len(self.points))
        u = self.points[i, :]

        # evolve it.
        scale = 1.
        accept = 0
        reject = 0
        ncall = 0
        while ncall < self.steps or accept == 0:
            new_u = self.propose_point(u, scale)
            new_v = self.prior_transform(new_u)
            new_logl = self.loglikelihood(new_v)
            if new_logl >= loglstar:
                u = new_u
                v = new_v
                logl = new_logl
                accept += 1
            else:
                reject += 1

            # adjust scale, aiming for acceptance ratio of 0.5.
            if accept > reject:
                scale *= math.exp(1. / accept)
            if accept < reject:
                scale /= math.exp(1. / reject)

            ncall += 1

        return u, v, logl, ncall


class SingleEllipsoidSampler(Sampler):
    """Bounds active points in a single ellipsoid and samples randomly
    from within that ellipsoid."""

    def set_options(self, options):
        self.enlarge = options.get('enlarge', 1.2)

    def update(self, pointvol):
        self.empty_queue()
        self.ell = bounding_ellipsoid(self.points, pointvol=pointvol,
                                      minvol=True)
        self.ell.scale_to_vol(self.ell.vol * self.enlarge)
        self.fill_queue()

    def propose_point(self):
        while True:
            u = self.ell.sample(rstate=self.rstate)
            if np.all(u > 0.) and np.all(u < 1.):
                break
        return u

    def new_point(self, loglstar):
        ncall = 0
        while True:
            u, v, logl = self.get_point_value()
            ncall += 1
            if logl >= loglstar:
                break

        return u, v, logl, ncall


class MultiEllipsoidSampler(Sampler):
    """Bounds active points in multiple ellipsoids and samples randomly
    from within joint distribution."""

    def set_options(self, options):
        self.enlarge = options.get('enlarge', 1.2)

    def update(self, pointvol):
        self.empty_queue()
        self.ells = bounding_ellipsoids(self.points, pointvol=pointvol)
        for ell in self.ells:
            ell.scale_to_vol(ell.vol * self.enlarge)
        self.fill_queue()

    def propose_point(self):
        while True:
            u = sample_ellipsoids(self.ells, rstate=self.rstate)
            if np.all(u > 0.) and np.all (u < 1.):
                break
        return u

    def new_point(self, loglstar):
        ncall = 0
        while True:
            u, v, logl = self.get_point_value()
            ncall += 1
            if logl >= loglstar:
                break

        return u, v, logl, ncall


# -----------------------------------------------------------------------------
# Main entry point

_SAMPLERS = {'classic': ClassicSampler,
             'single': SingleEllipsoidSampler,
             'multi': MultiEllipsoidSampler}

def sample(loglikelihood, prior_transform, ndim, npoints=100,
           method='single', update_interval=None, npdim=None,
           maxiter=None, maxcall=None, dlogz=None, decline_factor=None,
           rstate=None, callback=None, queue_size=None, pool=None, **options):
    """Perform nested sampling to evaluate Bayesian evidence.

    Parameters
    ----------
    loglikelihood : function
        Function returning log(likelihood) given parameters as a 1-d numpy
        array of length *ndim*.

    prior_transform : function
        Function translating a unit cube to the parameter space according to
        the prior. The input is a 1-d numpy array with length *ndim*, where
        each value is in the range [0, 1). The return value should also be a
        1-d numpy array with length *ndim*, where each value is a parameter.
        The return value is passed to the loglikelihood function. For example,
        for a 2 parameter model with flat priors in the range [0, 2), the
        function would be::

            def prior_transform(u):
                return 2.0 * u

    ndim : int
        Number of parameters returned by prior and accepted by loglikelihood.

    npoints : int, optional
        Number of active points. Larger numbers result in a more finely
        sampled posterior (more accurate evidence), but also a larger
        number of iterations required to converge. Default is 100.

    method : {'classic', 'single', 'multi'}, optional
        Method used to select new points. Choices are 'classic',
        single-ellipsoidal ('single'), multi-ellipsoidal ('multi'). Default
        is 'single'.

    update_interval : int, optional
        Only update the new point selector every ``update_interval``-th
        likelihood call. Update intervals larger than 1 can be more efficient
        when the likelihood function is very fast, particularly when
        using the multi-ellipsoid method. Default is round(0.6 * npoints).

    npdim : int, optional
        Number of parameters accepted by prior. This might differ from *ndim*
        in the case where a parameter of loglikelihood is dependent upon
        multiple independently distributed parameters, some of which may be
        nuisance parameters.

    maxiter : int, optional
        Maximum number of iterations. Iteration may stop earlier if
        termination condition is reached. Default is no limit.

    maxcall : int, optional
        Maximum number of likelihood evaluations. Iteration may stop earlier
        if termination condition is reached. Default is no limit.

    dlogz : float, optional
        If supplied, iteration will stop when the estimated contribution
        of the remaining prior volume to the total evidence falls below
        this threshold. Explicitly, the stopping criterion is
        ``log(z + z_est) - log(z) < dlogz`` where *z* is the current evidence
        from all saved samples, and *z_est* is the estimated contribution
        from the remaining volume. This option and decline_factor are
        mutually exclusive. If neither is specified, the default is
        ``dlogz=0.5``.

    decline_factor : float, optional
        If supplied, iteration will stop when the weight
        (likelihood times prior volume) of newly saved samples has been
        declining for ``decline_factor * nsamples`` consecutive samples.
        A value of 1.0 seems to work pretty well. This option and dlogz
        are mutually exclusive.

    rstate : `~numpy.random.RandomState`, optional
        RandomState instance. If not given, the global random state of the
        ``numpy.random`` module will be used.

    callback : function, optional
        Callback function to be called at each iteration. A single argument,
        a dictionary, is passed to the callback. The keys include ``'it'``,
        the current iteration number, and ``'logz'``, the current total
        log evidence of all saved points. To simply print these at each
        iteration, use the convience function
        ``callback=nestle.print_progress``.

    queue_size: int, optional
        Carry out evaluation in parallel by queueing up new active point 
        proposals using at most this many threads. Each thread independently
        proposes new live points until one is selected. 
        Default is no parallelism (queue_size=1). 

    pool: ThreadPoolExecutor
        Use this pool of workers to propose live points in parallel. If 
        queue_size>1 and `pool` is not specified, an Exception will be thrown. 


    Other Parameters
    ----------------
    steps : int, optional
        For the 'classic' method, the number of steps to take when selecting
        a new point. Default is 20.

    enlarge : float, optional
        For the 'single' and 'multi' methods, enlarge the ellipsoid(s) by
        this fraction in volume. Default is 1.2.


    Returns
    -------
    result : `Result`
        A dictionary-like object with attribute access: Attributes can be
        accessed with, for example, either ``result['niter']`` or
        ``result.niter``. Attributes:

        niter *(int)*
            Number of iterations.

        ncall *(int)*
            Number of likelihood calls.

        logz *(float)*
            Natural logarithm of evidence (integral of posterior).

        logzerr *(float)*
            Estimated numerical (sampling) error on *logz*.

        h *(float)*
            Information. This is a measure of the "peakiness" of the
            likelihood function. A constant likelihood has zero information.

        samples *(ndarray)*
            Parameter values of each sample. Shape is *(nsamples, ndim)*.

        logvol *(ndarray)*
            Natural log of prior volume of corresponding to each sample.
            Shape is *(nsamples,)*.

        logl *(ndarray)*
            Natural log of the likelihood for each sample, as returned by
            user-supplied *logl* function. Shape is *(nsamples,)*.

        weights *(ndarray)*
            Weight corresponding to each sample, normalized to unity.
            These are proportional to ``exp(logvol + logl)``. Shape is
            *(nsamples,)*.
    """

    if npdim is None:
        npdim = ndim

    if maxiter is None:
        maxiter = sys.maxsize

    if maxcall is None:
        maxcall = sys.maxsize

    if method == 'multi' and not HAVE_KMEANS:
        raise ValueError("scipy.cluster.vq.kmeans2 is required for the "
                         "'multi' method.")  # pragma: no cover

    if method not in _SAMPLERS:
        raise ValueError("Unknown method: {:r}".format(method))

    if npoints < 2 * ndim:
        warnings.warn("You really want to make npoints >= 2 * ndim!")

    if rstate is None:
        rstate = np.random

    # Stopping criterion.
    if dlogz is not None and decline_factor is not None:
        raise ValueError("Cannot specify two separate stopping criteria: "
                         "decline_factor and dlogz")
    elif dlogz is None and decline_factor is None:
        dlogz = 0.5

    if update_interval is None:
        update_interval = max(1, round(0.6 * npoints))
    else:
        update_interval = round(update_interval)
        if update_interval < 1:
            raise ValueError("update_interval must be >= 1")

    # Parallel evaluation.
    if queue_size is None or queue_size == 1:
        queue_size = 1
        pool = FakePool()
    else:
        if pool is None:
            raise ValueError("Missing pool. Please provide a Pool object.")

    # Initialize active points and calculate likelihoods
    active_u = rstate.rand(npoints, npdim)  # position in unit cube
    active_v = np.empty((npoints, ndim), dtype=np.float64)  # real params
    for i in range(npoints):
        active_v[i, :] = prior_transform(active_u[i, :])
    active_logl = np.fromiter(pool.map(loglikelihood, active_v), 
                              dtype=np.float64) # log likelihood
    sampler = _SAMPLERS[method](loglikelihood, prior_transform, active_u,
                                rstate, options, queue_size, pool)

    # Initialize values for nested sampling loop.
    saved_v = []  # stored points for posterior results
    saved_logl = []
    saved_logvol = []
    saved_logwt = []
    h = 0.0  # Information, initially 0.
    logz = -1e300  # ln(Evidence Z), initially Z=0.
    logvol = math.log(1.0 - math.exp(-1.0/npoints))  # first point removed will
                                                     # have volume 1-e^(1/n)
    ncall = npoints  # number of calls we already made

    # Initialize sampler
    sampler.update(1./npoints)

    callback_info = {'it': 0,
                     'logz': logz,
                     'active_u': active_u,
                     'sampler': sampler}

    # Nested sampling loop.
    ndecl = 0
    logwt_old = -np.inf
    it = 0
    since_update = 0
    while it < maxiter:
        if (callback is not None) and (it > 0):
            callback_info.update(it=it, logz=logz)
            callback(callback_info)

        # worst object in collection and its weight (= volume * likelihood)
        worst = np.argmin(active_logl)
        logwt = logvol + active_logl[worst]

        # update evidence Z and information h.
        logz_new = np.logaddexp(logz, logwt)
        h = (math.exp(logwt - logz_new) * active_logl[worst] +
             math.exp(logz - logz_new) * (h + logz) -
             logz_new)
        logz = logz_new

        # Add worst object to samples.
        saved_v.append(np.array(active_v[worst]))
        saved_logwt.append(logwt)
        saved_logvol.append(logvol)
        saved_logl.append(active_logl[worst])

        # The new likelihood constraint is that of the worst object.
        loglstar = active_logl[worst]

        expected_vol = math.exp(-it / npoints)
        pointvol = expected_vol / npoints

        # Update the sampler based on the current active points.
        if since_update >= update_interval:
            sampler.update(pointvol)
            since_update = 0

        # Choose a new point from within the likelihood constraint
        # (having logl > loglstar).
        u, v, logl, nc = sampler.new_point(loglstar)

        # replace worst point with new point
        active_u[worst] = u
        active_v[worst] = v
        active_logl[worst] = logl
        ncall += nc
        since_update += nc

        # Shrink interval
        logvol -= 1.0 / npoints

        # Stopping criterion 1: estimated fractional remaining evidence
        # below some threshold.
        if dlogz is not None:
            logz_remain = np.max(active_logl) - it / npoints
            if np.logaddexp(logz, logz_remain) - logz < dlogz:
                break

        # Stopping criterion 2: logwt has been declining for a while.
        if decline_factor is not None:
            ndecl = ndecl + 1 if logwt < logwt_old else 0
            logwt_old = logwt
            if ndecl > decline_factor * npoints:
                break

        if ncall > maxcall:
            break

        it += 1

    # Add remaining active points.
    # After N samples have been taken out, the remaining volume is
    # e^(-N/npoints). Thus, the remaining volume for each active point
    # is e^(-N/npoints) / npoints. The log of this for each object is:
    # log(e^(-N/npoints) / npoints) = -N/npoints - log(npoints)
    logvol = -len(saved_v) / npoints - math.log(npoints)
    for i in range(npoints):
        logwt = logvol + active_logl[i]
        logz_new = np.logaddexp(logz, logwt)
        h = (math.exp(logwt - logz_new) * active_logl[i] +
             math.exp(logz - logz_new) * (h + logz) -
             logz_new)
        logz = logz_new
        saved_v.append(np.array(active_v[i]))
        saved_logwt.append(logwt)
        saved_logl.append(active_logl[i])
        saved_logvol.append(logvol)

    # h should always be nonnegative (we take the sqrt below).
    # Numerical error makes it negative in pathological corner cases
    # such as flat likelihoods. Here we correct those cases to zero.
    if h < 0.0:
        if h > -SQRTEPS:
            h = 0.0
        else:
            raise RuntimeError("Negative h encountered (h={}). Please report "
                               "this as a likely bug.".format(h))

    return Result([
        ('niter', it + 1),
        ('ncall', ncall),
        ('logz', logz),
        ('logzerr', math.sqrt(h / npoints)),
        ('h', h),
        ('samples', np.array(saved_v)),
        ('weights', np.exp(np.array(saved_logwt) - logz)),
        ('logvol', np.array(saved_logvol)),
        ('logl', np.array(saved_logl))
        ])



#This script creates a frequency file from a FITS header. This is a helper
# script to make it easier to run RMsynth 1D or 3D. Run this first to generate
# the required frequency file. If you create a spectrum or cube from multiple
# FITS files, run it on the individual input files.
#This script assumes the FITS header has a FREQ axis, and that this axis
# accurately describes the frequency channels.

#version 1 by Boris Gbeasor, summer 2019
#ver2: modified by Cameron Van Eck



from astropy.io import fits
import numpy as np
import argparse


def save_freq_file():
    """
    Parses command line arguments, extracts FITS header, and saves it to a file.
    """
    # Parse the command line options
    parser = argparse.ArgumentParser(
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("fits_file",metavar="Infile.fits", nargs=1,
                        help="FITS cube with frequency axis.")
    parser.add_argument("freq_file", metavar="outfile.dat",
                        help="Name of the freq file to write.")

    args = parser.parse_args()


    freq_array = get_freq_array(args.fits_file[0])

    np.savetxt(args.freq_file, freq_array, delimiter="")
    print("Saving the frequencies list to {}".format(args.freq_file))

def get_fits_header(filename):
    hduList = fits.open(filename)
    header=hduList[0].header
    hduList.close()
    return header

def get_freq_array(filename):
    hduList = fits.open(filename)
    header = hduList[0].header
    for key, value in header.items():
        if type(value) is str:
            if value.lower() == "freq":
                n_axis = int(key[-1])

    iter_pixel= int(header["NAXIS{}".format(n_axis)]- header["CRPIX{}".format(n_axis)])
    freq_data = [header["CRVAL{}".format(n_axis)]]

    for i in range(0, iter_pixel):
        freq_data.append(freq_data[i] + header["CDELT{}".format(n_axis)])


    freq_data_array = np.asarray(freq_data)
    return freq_data_array



if __name__ == "__main__":
    save_freq_file()
